"""
Implementation of reaction network interface.
"""
from dataclasses import field
from typing import Iterable, List, Optional, Union

from graph_tool.all import Graph, Vertex, find_edge, find_vertex
from pymatgen.entries import Entry
from tqdm import tqdm

from rxn_network.core.cost_function import CostFunction
from rxn_network.core.network import Network
from rxn_network.costs.softplus import Softplus
from rxn_network.entries.experimental import ExperimentalReferenceEntry
from rxn_network.network.entry import NetworkEntry, NetworkEntryType
from rxn_network.network.gt import (
    initialize_graph,
    load_graph,
    save_graph,
    update_vertex_props,
    yens_ksp,
)
from rxn_network.pathways.basic import BasicPathway
from rxn_network.pathways.pathway_set import PathwaySet
from rxn_network.reactions.reaction_set import ReactionSet


class ReactionNetwork(Network):
    """
    Main reaction network class for building graphs of reactions and performing
    pathfinding.
    """

    def __init__(
        self,
        rxns: ReactionSet,
        cost_function: CostFunction = field(default_factory=Softplus),
    ):
        """
        Initialize a ReactionNetwork object for a set of reactions.

        Note: the precursors and target must be set by calling set_precursors() and
        set_target() respectively.

        Args:
            rxns: Reaction set of reactions
            enumerators: iterable of enumerators which will be called during the
                build of the network
            cost_function: the function used to calculate the cost of each reaction edge
            open_elem: Optional name of an element that is kept open during reaction
            chempot: Optional associated chemical potential of open element
        """
        super().__init__(rxns=rxns, cost_function=cost_function)

    def build(self):
        """
        Construct the reaction network graph object and store under the "graph"
        attribute. Does NOT initialize precursors or target; you must call
        set_precursors() or set_target() to do so.

        Returns:
            None
        """
        costs = self.rxns.calculate_costs(self.cost_function)
        self.logger.info("Building graph from reactions...")

        nodes, rxn_edges = get_rxn_nodes_and_edges(self.rxns)

        g = initialize_graph()
        g.add_vertex(len(nodes))

        for i, network_entry in enumerate(nodes):
            props = {"entry": network_entry, "type": network_entry.description.value}
            update_vertex_props(g, g.vertex(i), props)

        edge_list = []
        for edge, cost, rxn in zip(rxn_edges, costs, self.rxns):
            v1 = g.vertex(edge[0])
            v2 = g.vertex(edge[1])
            edge_list.append((v1, v2, cost, rxn, "reaction"))

        edge_list.extend(get_loopback_edges(g, nodes))

        g.add_edge_list(edge_list, eprops=[g.ep["cost"], g.ep["rxn"], g.ep["type"]])

        self._g = g

    def find_pathways(self, targets: List[str], k: float = 15) -> List[BasicPathway]:
        """
        Find the k-shortest paths to a provided list of 1 or more targets.

        Args:
            targets: List of the formulas of each target
            k: Number of shortest paths to find for each target

        Returns:
            List of BasicPathway objects to all provided targets.
        """
        if not self.precursors:
            raise AttributeError("Must call set_precursors() before pathfinding!")

        paths = []
        for target in targets:
            self.set_target(target)
            print(f"PATHS to {self.target.composition.reduced_formula} \n")
            print("--------------------------------------- \n")
            pathways = self._shortest_paths(k=k)
            paths.extend(pathways)

        paths = PathwaySet.from_paths(paths)

        return paths

    def set_precursors(self, precursors: Iterable[Union[Entry, str]]):
        """
        Sets the precursors of the network. Removes all references to previous
        precursors.

        If entries are provided, will use the entries to set the precursors. If strings
        are provided, will automatically find minimum-energy entries with matching
        reduced_formula.

        Args:
            precursors: iterable of

        Returns:
            None
        """
        g = self._g
        if not g:
            raise ValueError("Must call build() before setting precursors!")

        precursors = {
            p
            if isinstance(p, (Entry, ExperimentalReferenceEntry))
            else self.entries.get_min_entry_by_formula(p)
            for p in precursors
        }

        if precursors == self.precursors:
            return

        if self.precursors:
            precursors_v = find_vertex(
                g, g.vp["type"], NetworkEntryType.Precursors.value
            )[0]
            g.remove_vertex(precursors_v)
            loopback_edges = find_edge(g, g.ep["type"], "loopback_precursors")
            for e in loopback_edges:
                g.remove_edge(e)
        elif not all(p in self.entries for p in precursors):
            raise ValueError("One or more precursors are not included in network!")

        precursors_v = g.add_vertex()
        precursors_entry = NetworkEntry(precursors, NetworkEntryType.Precursors)

        props = {"entry": precursors_entry, "type": precursors_entry.description.value}
        update_vertex_props(g, precursors_v, props)

        add_edges = []
        for v in g.vertices():
            entry = g.vp["entry"][v]
            if not entry:
                continue
            if entry.description.value == NetworkEntryType.Reactants.value:
                if entry.entries.issubset(precursors):
                    add_edges.append((precursors_v, v, 0.0, None, "precursors"))
            elif entry.description.value == NetworkEntryType.Products.value:
                for v2 in g.vertices():
                    entry2 = g.vp["entry"][v2]
                    if entry2.description.value == NetworkEntryType.Reactants.value:
                        if precursors.issuperset(entry2.entries):
                            continue
                        if precursors.union(entry.entries).issuperset(entry2.entries):
                            add_edges.append((v, v2, 0.0, None, "loopback_precursors"))

        g.add_edge_list(add_edges, eprops=[g.ep["cost"], g.ep["rxn"], g.ep["type"]])

        self._precursors = precursors

    def set_target(self, target: Union[Entry, str]):
        """
        If entry is provided, will use that entry to set the target. If string is
        provided, will automatically find minimum-energy entry with matching
        reduced_formula.

        Args:
            target: Entry, or string of reduced formula, of target

        Returns:
            None
        """
        g = self._g
        if not g:
            raise ValueError("Must call build() before setting target!")

        target = (
            target
            if isinstance(target, (Entry, ExperimentalReferenceEntry))
            else self.entries.get_min_entry_by_formula(target)
        )

        if target == self.target:
            return

        if target not in self.entries:
            raise ValueError("Target is not included in network!")

        if self.target:
            target_v = find_vertex(g, g.vp["type"], NetworkEntryType.Target.value)[0]
            g.remove_vertex(target_v)

        target_v = g.add_vertex()
        target_entry = NetworkEntry([target], NetworkEntryType.Target)
        props = {"entry": target_entry, "type": target_entry.description.value}
        update_vertex_props(g, target_v, props)

        add_edges = []
        for v in g.vertices():
            entry = g.vp["entry"][v]
            if not entry:
                continue
            if entry.description.value != NetworkEntryType.Products.value:
                continue
            if target in entry.entries:
                add_edges.append([v, target_v, 0.0, None, "target"])

        g.add_edge_list(add_edges, eprops=[g.ep["cost"], g.ep["rxn"], g.ep["type"]])

        self._target = target

    def load_graph(self, filename: str):
        """
        Loads graph-tool graph from file.

        Args:
            filename: Filename of graph object to load (for example, .gt or .gt.gz
                format)

        Returns: None
        """
        self._g = load_graph(filename)

    def write_graph(self, filename: Optional[str] = None):
        """
        Writes graph to file. If filename is not provided, will write to CHEMSYS.gt.gz
        (where CHEMSYS is the chemical system of the network)

        Args:
            filename: Filename to write to. If None, writes to default filename.

        Returns: None
        """
        if not filename:
            filename = f"{self.chemsys}.gt.gz"

        save_graph(self._g, filename)

    def _shortest_paths(self, k):
        """Finds the k shortest paths using Yen's algorithm and returns BasicPathways"""
        g = self._g
        paths = []

        precursors_v = find_vertex(g, g.vp["type"], NetworkEntryType.Precursors.value)[
            0
        ]
        target_v = find_vertex(g, g.vp["type"], NetworkEntryType.Target.value)[0]

        for path in yens_ksp(g, k, precursors_v, target_v):
            paths.append(self._path_from_graph(g, path))

        for path in paths:
            print(path, "\n")

        return paths

    @staticmethod
    def _path_from_graph(g, path):
        """Gets a BasicPathway object from a shortest path found in the network"""
        rxns = []
        costs = []

        for step, v in enumerate(path):
            if g.vp["type"][v] == NetworkEntryType.Products.value:
                e = g.edge(path[step - 1], v)

                rxns.append(g.ep["rxn"][e])
                costs.append(g.ep["cost"][e])

        return BasicPathway(reactions=rxns, costs=costs)

    @classmethod
    def from_dict_and_file(cls, d: dict, filename: str):
        """
        Convenience constructor method that loads a ReactionNetwork object from a
        dictionary (MSONable version) and a filename (to load graph object in
        graph-tool).

        Args:
            d: Dictionary containing the ReactionNetwork object
            filename: Filename of graph object to load (for example, .gt or .gt.gz
                format)

        Returns:
            ReactionNetwork object with loaded graph
        """
        rn = cls.from_dict(d)
        rn.load_graph(filename)  # pylint: disable=no-member

        return rn

    @property
    def graph(self):
        """Returns the network object in graph-tool"""
        return self._g

    @property
    def chemsys(self) -> str:
        """Returns a string of the chemical system of the network"""
        return "-".join(sorted(self.entries.chemsys))

    def as_dict(self) -> dict:
        """Return MSONable dict"""
        d = super().as_dict()
        d["precursors"] = list(self.precursors) if self.precursors else None
        d["target"] = self.target
        return d

    @classmethod
    def from_dict(cls, d):
        """Instantiate object from MSONable dict"""
        precursors = d.pop("precursors", None)
        target = d.pop("target", None)

        rn = super().from_dict(d)
        rn._precursors = precursors  # pylint: disable=protected-access
        rn._target = target  # pylint: disable=protected-access

        return rn

    def __repr__(self):
        return (
            "ReactionNetwork for chemical system: "
            f"{self.chemsys}, "
            f"with Graph: {str(self._g)}"
        )


def get_rxn_nodes_and_edges(rxns: ReactionSet):
    """
    Given a reaction set, return a list of nodes and edges for constructing the
    reaction network.

    Args:
        rxns: a list of enumerated ComputedReaction objects to build a network from.

    Returns:
        A tuple consisting of (nodes, edges) where nodes is a list of NetworkEntry
        objects and edges is a list of tuples of the form (source_idx, target_idx).
    """
    nodes, edges = [], []

    for rxn in tqdm(rxns):
        reactant_node = NetworkEntry(rxn.reactant_entries, NetworkEntryType.Reactants)
        product_node = NetworkEntry(rxn.product_entries, NetworkEntryType.Products)

        if reactant_node not in nodes:
            nodes.append(reactant_node)
            reactant_idx = len(nodes) - 1
        else:
            reactant_idx = nodes.index(reactant_node)

        if product_node not in nodes:
            nodes.append(product_node)
            product_idx = len(nodes) - 1
        else:
            product_idx = nodes.index(product_node)

        edges.append((reactant_idx, product_idx))

    return nodes, edges


def get_loopback_edges(g: Graph, nodes: List[Vertex]):
    """
    Given a graph and a list of nodes to check, this function finds and returns loopback
    edges (i.e., edges that connect a product node to its equivalent reactant node)

    Args:
        g: graph-tool Graph object
        nodes: List of vertices from which to find loopback edges

    Returns:
        A list of tuples of the form (source_idx, target_idx, cost=0, rxn=None,
        type="loopback")
    """
    edges = []
    for idx1, p in enumerate(nodes):
        if p.description.value != NetworkEntryType.Products.value:
            continue
        for idx2, r in enumerate(nodes):
            if r.description.value != NetworkEntryType.Reactants.value:
                continue
            if p.entries == r.entries:
                edges.append((g.vertex(idx1), g.vertex(idx2), 0.0, None, "loopback"))

    return edges



"""
This module implements two types of basic reaction enumerators, differing in the option
to consider open entries.
"""
from copy import deepcopy
from itertools import combinations, product
from math import comb
from typing import List, Optional, Set

import ray
from pymatgen.analysis.phase_diagram import GrandPotentialPhaseDiagram, PhaseDiagram
from pymatgen.entries.computed_entries import ComputedEntry
from tqdm import tqdm

from rxn_network.core.enumerator import Enumerator
from rxn_network.entries.entry_set import GibbsEntrySet
from rxn_network.entries.utils import initialize_entry
from rxn_network.enumerators.utils import get_rxn_info, group_by_chemsys
from rxn_network.reactions.computed import ComputedReaction
from rxn_network.reactions.reaction_set import ReactionSet
from rxn_network.utils.funcs import get_logger, grouper, limited_powerset
from rxn_network.utils.ray import initialize_ray

logger = get_logger(__name__)


class BasicEnumerator(Enumerator):
    """
    Enumerator for finding all simple reactions within a set of entries, up to a
    maximum reactant/product cardinality (n); i.e., how many phases on either side of
    the reaction. This approach does not explicitly take into account thermodynamic
    stability (i.e. phase diagram). This allows for enumeration of reactions where the
    products may not be stable with respect to each other.
    """

    CHUNK_SIZE = 2500

    def __init__(
        self,
        precursors: Optional[List[str]] = None,
        targets: Optional[List[str]] = None,
        n: int = 2,
        exclusive_precursors: bool = True,
        exclusive_targets: bool = False,
        filter_by_chemsys: Optional[str] = None,
        max_num_constraints=1,
        remove_unbalanced: bool = True,
        remove_changed: bool = True,
        calculate_e_above_hulls: bool = False,
        quiet: bool = False,
    ):
        """
        Supplied target and calculator parameters are automatically initialized as
        objects during enumeration.

        Args:
            precursors: Optional list of precursor formulas; only reactions
                which contain at least these phases as reactants will be enumerated. See
                the "exclusive_precursors" parameter for more details.
            targets: Optional list of target formulas; only reactions which include
                formation of at least one of these targets will be enumerated. See the
                "exclusive_targets" parameter for more details.
            n: Maximum reactant/product cardinality; i.e., largest possible number of
                entries on either side of the reaction. Defaults to 2.
            exclusive_precursors: Whether to consider only reactions that have reactants
                which are a subset of the provided list of precursors. In other
                words, if True, this only identifies reactions with reactants selected
                from the precursors argument. Defaults to True.
            exclusive_targets: Whether to consider only reactions that make the
                form products that are a subset of the provided list of targets. If
                False, this only identifies reactions with no unspecified byproducts.
                Defualts to False.
            remove_unbalanced: Whether to remove reactions which are unbalanced.
                Defaults to True.
            remove_changed: Whether to remove reactions which can only be balanced by
                removing a reactant/product or having it change sides. Defaults to True.
            calculate_e_above_hulls: Whether to calculate e_above_hull for each entry
                upon initialization of the entries at the beginning of enumeration.
            quiet: Whether to run in quiet mode (no progress bar). Defaults to False.
        """

        super().__init__(precursors=precursors, targets=targets)

        self.n = n
        self.exclusive_precursors = exclusive_precursors
        self.exclusive_targets = exclusive_targets
        self.filter_by_chemsys = filter_by_chemsys
        self.max_num_constraints = max_num_constraints
        self.remove_unbalanced = remove_unbalanced
        self.remove_changed = remove_changed
        self.calculate_e_above_hulls = calculate_e_above_hulls
        self.quiet = quiet

        self._stabilize = False

        self._p_set_func = "issuperset" if self.exclusive_precursors else "intersection"
        self._t_set_func = "issuperset" if self.exclusive_targets else "intersection"

        self.open_phases: Optional[List] = None
        self._build_pd = False
        self._build_grand_pd = False

    def enumerate(self, entries: GibbsEntrySet, batch_size=None) -> ReactionSet:
        """
        Calculate all possible reactions given a set of entries. If the enumerator was
        initialized with specified precursors or target, the reactions will be filtered
        by these constraints. Every enumerator follows a standard procedure:

        1. Initialize entries, i.e., ensure that precursors and target are considered
        stable entries within the entry set. If using ChempotDistanceCalculator, ensure
        that entries are filtered by stability.

        2. Get a dictionary representing every possible "node", i.e. phase combination,
        grouped by chemical system.

        3. Filter the combos dictionary for chemical systems which are not relevant;
        i.e., don't contain elements in precursors and/or target.

        4. Iterate through each chemical system, initializing calculators,
        and computing all possible reactions for reactant/product pair and/or
        thermodynamically predicted reactions for given reactants.

        5. Add reactions to growing list, repeat Step 4 until combos dict exhausted.

        Args:
            entries: the set of all entries to enumerate from
        """

        initialize_ray()

        entries, precursors, targets, open_entries = self._get_initialized_entries(
            entries
        )

        combos_dict = self._get_combos_dict(
            entries,
            precursors,
            targets,
            open_entries,
        )
        open_combos = self._get_open_combos(open_entries)

        if not open_combos:
            open_combos = []

        items = combos_dict.items()

        precursors = ray.put(precursors)
        targets = ray.put(targets)
        react_function = ray.put(self._react_function)
        open_entries = ray.put(open_entries)
        p_set_func = ray.put(self._p_set_func)
        t_set_func = ray.put(self._t_set_func)
        remove_unbalanced = ray.put(self.remove_unbalanced)
        remove_changed = ray.put(self.remove_changed)
        max_num_constraints = ray.put(self.max_num_constraints)

        rxn_chunk_refs = []  # type: ignore
        results = []

        if not batch_size:
            batch_size = ray.cluster_resources()["CPU"] * 2

        with tqdm(
            total=self._num_chunks(items, open_combos), disable=self.quiet
        ) as pbar:
            for item in items:
                chemsys, combos = item

                elems = chemsys.split("-")

                filtered_entries = None
                pd = None
                grand_pd = None

                if self.build_pd or self.build_grand_pd:
                    filtered_entries = entries.get_subset_in_chemsys(elems)

                if self.build_pd:
                    pd = PhaseDiagram(filtered_entries)

                if self.build_grand_pd:
                    chempots = getattr(self, "chempots")
                    grand_pd = GrandPotentialPhaseDiagram(filtered_entries, chempots)

                filtered_entries = ray.put(filtered_entries)
                pd = ray.put(pd)
                grand_pd = ray.put(grand_pd)

                for rxn_iterable_chunk in grouper(
                    self._get_rxn_iterable(combos, open_combos), self.CHUNK_SIZE
                ):
                    if len(rxn_chunk_refs) > batch_size:
                        num_ready = len(rxn_chunk_refs) - batch_size
                        newly_completed, rxn_chunk_refs = ray.wait(
                            rxn_chunk_refs, num_returns=num_ready
                        )
                        for completed_ref in newly_completed:
                            results.extend(ray.get(completed_ref))

                            pbar.update(1)

                    rxn_chunk_refs.append(
                        _react.remote(
                            rxn_iterable_chunk,
                            react_function,
                            open_entries,
                            precursors,
                            targets,
                            p_set_func,
                            t_set_func,
                            remove_unbalanced,
                            remove_changed,
                            max_num_constraints,
                            filtered_entries,
                            pd,
                            grand_pd,
                        )
                    )

            newly_completed, rxn_chunk_refs = ray.wait(
                rxn_chunk_refs, num_returns=len(rxn_chunk_refs)
            )
            for completed_ref in newly_completed:
                results.extend(ray.get(completed_ref))
                pbar.update(1)

        all_indices, all_coeffs, all_data = [], [], []
        for r in results:
            all_indices.append(r[0])
            all_coeffs.append(r[1])
            all_data.append(r[2])

        rxn_set = ReactionSet(
            entries.entries_list, all_indices, all_coeffs, all_data=all_data
        )
        rxn_set = rxn_set.filter_duplicates()

        return rxn_set

    @classmethod
    def _num_chunks(cls, items, open_combos):
        _ = open_combos  # not used

        n = 0
        for _, i in items:
            num_combos = cls._rxn_iter_length(i, open_combos)
            if num_combos > 0:
                n += num_combos // cls.CHUNK_SIZE + 1

        return n

    @staticmethod
    def _rxn_iter_length(combos, open_combos):
        _ = open_combos  # not used
        return comb(len(combos), 2)

    def _get_combos_dict(
        self, entries, precursor_entries, target_entries, open_entries
    ):
        """
        Gets all possible entry combinations up to predefined cardinality (n), filtered
        and grouped by chemical system.
        """
        precursor_elems = [
            [str(el) for el in e.composition.elements] for e in precursor_entries
        ]
        target_elems = [
            [str(el) for el in e.composition.elements] for e in target_entries
        ]
        all_open_elems = {el for e in open_entries for el in e.composition.elements}

        entries = entries - open_entries

        combos = [set(c) for c in limited_powerset(entries, self.n)]
        combos_dict = group_by_chemsys(combos, all_open_elems)

        filtered_combos = self._filter_dict_by_elems(
            combos_dict,
            precursor_elems,
            target_elems,
            all_open_elems,
        )

        return filtered_combos

    def _get_open_combos(  # pylint: disable=useless-return
        self, open_entries
    ) -> Optional[List[Set[ComputedEntry]]]:
        """No open entries for BasicEnumerator, returns None"""
        _ = (self, open_entries)  # unused_arguments
        return None

    @staticmethod
    def _react_function(reactants, products, **kwargs):
        _ = kwargs  # unused_argument
        forward_rxn = ComputedReaction.balance(reactants, products)
        backward_rxn = forward_rxn.reverse()
        return [forward_rxn, backward_rxn]

    @staticmethod
    def _get_rxn_iterable(combos, open_combos):
        """Get all reaction/product combinations"""
        _ = open_combos  # unused argument

        return combinations(combos, 2)

    def _get_initialized_entries(self, entries):
        """Returns initialized entries, precursors, target, and open entries"""

        def initialize_entries_list(ents):
            new_ents = {initialize_entry(f, entries, self.stabilize) for f in ents}
            return new_ents

        precursors, targets = set(), set()

        entries_new = GibbsEntrySet(
            deepcopy(entries), calculate_e_above_hulls=self.calculate_e_above_hulls
        )

        if self.precursors:
            precursors = initialize_entries_list(self.precursors)
        if self.targets:
            targets = initialize_entries_list(self.targets)

        for e in precursors | targets:
            if e not in entries_new:
                try:
                    old_e = entries_new.get_min_entry_by_formula(
                        e.composition.reduced_formula
                    )
                    entries_new.discard(old_e)
                except KeyError:
                    pass

                entries_new.add(e)

        if self.stabilize:
            entries_new = entries_new.filter_by_stability(e_above_hull=0.0)
            logger.info("Filtering by stable entries!")

        entries_new.build_indices()

        open_entries = set()
        if self.open_phases:
            open_entries = {
                e
                for e in entries_new
                if e.composition.reduced_formula in self.open_phases
            }

        return entries_new, precursors, targets, open_entries

    def _filter_dict_by_elems(
        self,
        combos_dict,
        precursor_elems,
        target_elems,
        all_open_elems,
    ):
        """Filters the dictionary of combinations by elements"""
        filtered_dict = {}

        all_precursor_elems = {el for g in precursor_elems for el in g}
        all_target_elems = {el for g in target_elems for el in g}
        all_open_elems = {str(el) for el in all_open_elems}

        filter_elems = None
        if self.filter_by_chemsys:
            filter_elems = set(self.filter_by_chemsys.split("-"))

        for chemsys, combos in combos_dict.items():
            elems = set(chemsys.split("-"))

            if filter_elems:
                if not elems.issuperset(filter_elems):
                    continue

            if len(elems) >= 10 or len(elems) == 1:  # too few or too many elements
                continue

            if precursor_elems:
                if not getattr(all_precursor_elems | all_open_elems, self._p_set_func)(
                    elems
                ):
                    continue

            if target_elems:
                if not getattr(all_target_elems | all_open_elems, self._t_set_func)(
                    elems
                ):
                    continue

            filtered_dict[chemsys] = combos

        return filtered_dict

    @property
    def stabilize(self):
        """Whether or not to use only stable entries in analysis"""
        return self._stabilize

    @property
    def build_pd(self):
        """Whether or not to build a PhaseDiagram object during reaction enumeration (
        useful for some analyses)"""
        return self._build_pd

    @property
    def build_grand_pd(self):
        """Whether or not to build a GrandPotentialPhaseDiagram object during
        reaction enumeration (useful for some analyses)"""
        return self._build_grand_pd


class BasicOpenEnumerator(BasicEnumerator):
    """
    Enumerator for finding all simple reactions within a set of entries, up to a
    maximum reactant/product cardinality (n), with any number of open phases. Note:
    this does not return OpenComputedReaction objects (this can be calculated using
    the ReactionSet class).
    """

    CHUNK_SIZE = 2500

    def __init__(
        self,
        open_phases: List[str],
        precursors: Optional[List[str]] = None,
        targets: Optional[List[str]] = None,
        n: int = 2,
        exclusive_precursors: bool = True,
        exclusive_targets: bool = False,
        filter_by_chemsys: Optional[str] = None,
        max_num_constraints: int = 1,
        remove_unbalanced: bool = True,
        remove_changed: bool = True,
        calculate_e_above_hulls: bool = False,
        quiet: bool = False,
    ):
        """
        Supplied target and calculator parameters are automatically initialized as
        objects during enumeration.

        Args:
            open_phases: List of formulas of open entries (e.g. ["O2"])
            precursors: Optional list of formulas of precursor phases; only reactions
                which have these phases as reactants will be enumerated.
            targets: Optional list of formulas of targets; only reactions
                which make this target will be enumerated.
            calculators: Optional list of Calculator object names; see calculators
                module for options (e.g., ["ChempotDistanceCalculator]).
            n: Maximum reactant/product cardinality; i.e., largest possible number of
                entries on either side of the reaction.
            exclusive_precursors: Whether to consider only reactions that have
                reactants which are a subset of the provided list of precursors.
                Defaults to True.
            exclusive_targets: Whether to consider only reactions that make the
                provided target directly (i.e. with no byproducts). Defualts to False.
            remove_unbalanced: Whether to remove reactions which are unbalanced.
                Defaults to True
            remove_changed: Whether to remove reactions which can only be balanced by
                removing a reactant/product or having it change sides. Defaults to True.
            quiet: Whether to run in quiet mode (no progress bar). Defaults to False.
        """
        super().__init__(
            precursors=precursors,
            targets=targets,
            n=n,
            exclusive_precursors=exclusive_precursors,
            exclusive_targets=exclusive_targets,
            filter_by_chemsys=filter_by_chemsys,
            max_num_constraints=max_num_constraints,
            remove_unbalanced=remove_unbalanced,
            remove_changed=remove_changed,
        )
        self.open_phases: List[str] = open_phases

    @staticmethod
    def _rxn_iter_length(combos, open_combos):
        num_combos_with_open = sum(
            1 if not i & j else 0 for i in combos for j in open_combos
        )

        return len(combos) * num_combos_with_open

    def _get_open_combos(self, open_entries):
        """Get all possible combinations of open entries. For a single entry,
        this is just the entry itself."""
        open_combos = [
            set(c) for c in limited_powerset(open_entries, len(open_entries))
        ]
        return open_combos

    @staticmethod
    def _get_rxn_iterable(combos, open_combos):
        """Get all reaction/product combinations."""
        combos_with_open = [
            combo | open_combo
            for combo in combos
            for open_combo in open_combos
            if not combo & open_combo
        ]
        rxn_iter = product(combos, combos_with_open)

        return rxn_iter


@ray.remote
def _react(
    rxn_iterable,
    react_function,
    open_entries,
    precursors,
    targets,
    p_set_func,
    t_set_func,
    remove_unbalanced,
    remove_changed,
    max_num_constraints,
    filtered_entries,
    pd,
    grand_pd,
):
    """
    This function is a wrapper for the specific react function of each enumerator. This
    wrapper contains the logic used for filtering out reactions based on the
    user-defined enumerator settings. It can also be called as a remote function using
    ray, allowing for parallel computation during reaction enumeration.

    Note: this function is not intended to to be called directly!

    """
    all_rxns = []

    for rp in rxn_iterable:
        if not rp:
            continue

        r = set(rp[0]) if rp[0] else set()
        p = set(rp[1]) if rp[1] else set()

        all_phases = r | p

        precursor_func = (
            getattr(precursors | open_entries, p_set_func)
            if precursors
            else lambda e: True
        )
        target_func = (
            getattr(targets | open_entries, t_set_func) if targets else lambda e: True
        )

        if (
            (r & p)
            or (precursors and not precursors & all_phases)
            or (p and targets and not targets & all_phases)
        ):
            continue

        if not (precursor_func(r) or (p and precursor_func(p))):
            continue
        if p and not (target_func(r) or target_func(p)):
            continue

        suggested_rxns = react_function(
            r, p, filtered_entries=filtered_entries, pd=pd, grand_pd=grand_pd
        )

        rxns = []
        for rxn in suggested_rxns:
            if (
                rxn.is_identity
                or (remove_unbalanced and not rxn.balanced)
                or (remove_changed and rxn.lowest_num_errors != 0)
                or rxn.data["num_constraints"] > max_num_constraints
            ):
                continue

            reactant_entries = set(rxn.reactant_entries) - open_entries
            product_entries = set(rxn.product_entries) - open_entries

            if precursor_func(reactant_entries) and target_func(product_entries):
                rxns.append(get_rxn_info(rxn))

        all_rxns.extend(rxns)

    return all_rxns



"""
Basic interface for a cost function
"""
from abc import ABCMeta, abstractmethod

from monty.json import MSONable

from rxn_network.core.reaction import Reaction


class CostFunction(MSONable, metaclass=ABCMeta):
    """
    Base definition for a cost function
    """

    @abstractmethod
    def evaluate(self, rxn: Reaction) -> float:
        """
        Evaluates the total cost function on a reaction
        """



"""
Graph-related functions specific to the graph-tool library. Used in the network module.
"""
from queue import Empty, PriorityQueue
from typing import Dict

from graph_tool import Graph, GraphView, Vertex
from graph_tool import load_graph as load_g
from graph_tool.topology import shortest_path

DEFAULT_VERTEX_PROPS = {"entry": "object", "type": "int"}
DEFAULT_EDGE_PROPS = {"rxn": "object", "cost": "double", "type": "string"}


def yens_ksp(
    g: Graph,
    num_k: int,
    precursors_v: Vertex,
    target_v: Vertex,
    edge_prop: str = "bool",
    cost_prop: str = "cost",
):
    """
    Yen's Algorithm for k-shortest paths, adopted for graph-tool. Utilizes GraphView
    objects to speed up filtering. I

    This implementation was inspired by the igraph implementation by Antonin Lenfant.

    Reference:
        Jin Y. Yen, "Finding the K Shortest Loopless Paths n a Network", Management
        Science, Vol. 17, No. 11, Theory Series (Jul., 1971), pp. 712-716.

    Args:
        g: the graph-tool graph object.
        num_k: number of k shortest paths that should be found.
        precursors_v: graph-tool vertex object containing precursors.
        target_v: graph-tool vertex object containing target.
        edge_prop: name of edge property map which allows for filtering edges.
            Defaults to the word "bool".
        cost_prop: name of edge property map that stores edge weights/costs.
            Defaults to the word "weight".
    Returns:
        List of lists of graph vertices corresponding to each shortest path
            (sorted in increasing order by cost).
    """

    def path_cost(vertices):
        """Calculates path cost given a list of vertices."""
        cost = 0
        for j in range(len(vertices) - 1):
            cost += g.ep[cost_prop][g.edge(vertices[j], vertices[j + 1])]
        return cost

    path = shortest_path(g, precursors_v, target_v, weights=g.ep[cost_prop])[0]

    if not path:
        return []
    a = [path]
    a_costs = [path_cost(path)]

    b = PriorityQueue()  # type: ignore

    g.ep["bool"] = g.new_edge_property("bool", val=True)

    for k in range(1, num_k):
        try:
            prev_path = a[k - 1]
        except IndexError:
            print(f"Identified only k={k-1} paths before exiting. \n")
            break

        for i in range(len(prev_path) - 1):
            spur_v = prev_path[i]
            root_path = prev_path[:i]

            filtered_edges = []

            for path in a:
                if len(path) - 1 > i and root_path == path[:i]:
                    e = g.edge(path[i], path[i + 1])
                    if not e:
                        continue
                    g.ep[edge_prop][e] = False
                    filtered_edges.append(e)

            gv = GraphView(g, efilt=g.ep[edge_prop])
            spur_path = shortest_path(gv, spur_v, target_v, weights=g.ep[cost_prop])[0]

            for e in filtered_edges:
                g.ep[edge_prop][e] = True

            if spur_path:
                total_path = root_path + spur_path
                total_path_cost = path_cost(total_path)
                b.put((total_path_cost, total_path))

        while True:
            try:
                cost_, path_ = b.get(block=False)
            except Empty:
                break
            if path_ not in a:
                a.append(path_)
                a_costs.append(cost_)
                break

    return a


def initialize_graph(
    vertex_props: Dict[str, str] = None, edge_props: Dict[str, str] = None
) -> Graph:
    """
    Initializes a graph-tool graph with the provided vertex and edge properties,
    provided a dictionary of the form {"prop": "type"}, where prop is the name of the
    property and type is the data type.

    Args:
        vertex_props: Dictionary of vertex property map properties in the form of
            {"prop": "type"}.
        edge_props: Dictionary of edge property map properties in the form of
            {"prop": "type"}.

    Returns:
        Graph: a graph-tool graph object.
    """
    g = Graph()

    if not vertex_props:
        vertex_props = {}
    if not edge_props:
        edge_props = {}

    vertex_props.update(DEFAULT_VERTEX_PROPS)
    edge_props.update(DEFAULT_EDGE_PROPS)

    for name, obj_type in vertex_props.items():
        g.vp[name] = g.new_vertex_property(obj_type)
    for name, obj_type in edge_props.items():
        g.ep[name] = g.new_edge_property(obj_type)

    return g


def load_graph(filename: str):
    """
    Loads a graph-tool graph from a file name.
    """
    g = load_g(filename)
    return g


def save_graph(g: Graph, filename: str):
    """
    Saves a graph-tool graph to a file by calling the Graph.save(filename) method.
    """
    g.save(filename)


def update_vertex_props(g, v, prop_dict):
    """
    Helper method for updating several vertex properties at once in a graph-tool
    Graph; note that this updates the vertex property map in place.

    Args:
        g (gt.Graph): a graph-tool Graph object.
        v (gt.Vertex or int): a graph-tool Vertex object (or its index) for a vertex
            in the provided graph.
        prop_dict (dict): a dictionary of the form {"prop": val}, where prop is the
            name of a VertexPropertyMap of the graph and val is the new updated
            value for that vertex's property.

    Returns:
        None
    """
    for prop, val in prop_dict.items():
        g.vp[prop][v] = val



"""
Basic interface for a reaction network.
"""
import logging
from abc import ABCMeta, abstractmethod
from typing import List

from monty.json import MSONable

from rxn_network.core.cost_function import CostFunction
from rxn_network.core.pathway import Pathway
from rxn_network.entries.entry_set import GibbsEntrySet
from rxn_network.reactions.reaction_set import ReactionSet


class Network(MSONable, metaclass=ABCMeta):
    """
    Base definition for a reaction network.
    """

    def __init__(
        self,
        rxns: ReactionSet,
        cost_function: CostFunction,
    ):
        self.logger = logging.getLogger(str(self.__class__.__name__))
        self.logger.setLevel("INFO")
        self.rxns = rxns
        self.cost_function = cost_function

        self.entries = GibbsEntrySet(rxns.entries)
        self.entries.build_indices()

        self._g = None
        self._precursors = None
        self._target = None

    @abstractmethod
    def build(self):
        """Construct the network from the supplied enumerators"""

    @abstractmethod
    def find_pathways(self, target, k) -> List[Pathway]:
        """Find reaction pathways"""

    @abstractmethod
    def set_precursors(self, precursors):
        """Set the phases used as precursors in the network"""

    @abstractmethod
    def set_target(self, target):
        """Set the phase used as a target in the network"""

    @property
    def precursors(self):
        """The phases used as precursors in the network"""
        return self._precursors

    @property
    def target(self):
        """The phase used as a target in the network"""
        return self._target

    @property
    def graph(self):
        """Returns the network object in graph-tool"""
        return self._g

    @property
    def chemsys(self):
        """A string representing the chemical system (elements) of the network"""
        return "-".join(sorted(self.entries.chemsys))



"""
Basic interface for a reaction pathway solver.
"""
import logging
from abc import ABCMeta
from typing import List

from monty.json import MSONable

from rxn_network.core.pathway import Pathway
from rxn_network.core.reaction import Reaction


class Solver(MSONable, metaclass=ABCMeta):
    """
    Base definition for a pathway solver class.
    """

    def __init__(self, pathways):
        self.logger = logging.getLogger(str(self.__class__.__name__))
        self.logger.setLevel("INFO")

        self._pathways = pathways

        rxns = []
        costs = []

        for path in self._pathways.get_paths():
            for rxn, cost in zip(path.reactions, path.costs):
                if rxn not in rxns:
                    rxns.append(rxn)
                    costs.append(cost)

        self._reactions = rxns
        self._costs = costs

    @property
    def pathways(self) -> List[Pathway]:
        """Pathways used in solver class"""
        return self._pathways

    @property
    def reactions(self) -> List[Reaction]:
        """Reactions used in solver class"""
        return self._reactions

    @property
    def costs(self) -> List[float]:
        """Costs used in solver class"""
        return self._costs

    @property
    def num_rxns(self) -> int:
        """Length of the reaction list"""
        return len(self.reactions)

    @property
    def num_entries(self) -> int:
        """Length of entry list"""
        return len(self._entries)



"""
Functions for visualizing/plotting reaction networks.
"""
import warnings
from typing import Optional

import graph_tool.all as gt
import matplotlib.cm
import numpy as np


def plot_network(
    graph: gt.Graph,
    vertex_cmap_name: str = "jet",
    edge_cmap_name: str = "PuBuGn_r",
    output: Optional[str] = None,
    cost_pos_scale_factor: float = 10.0,
):
    """
    Plots a reaction network using graph-tool visualization tools (i.e., graph_draw())

    Args:
        graph: a graph-tool Graph object
        vertex_cmap_name: the name of . Defaults to "jet".
        edge_cmap_name: Defaults to "PuBuGn_r".
        output: Optional output filename
        cost_pos_scale_factor

    """
    g = graph.copy()

    costs = np.array(g.ep["cost"].get_array().tolist())

    edge_weights = gt.prop_to_size(
        g.new_edge_property("float", cost_pos_scale_factor * costs), mi=0.1, ma=35
    )

    deg = gt.prop_to_size(g.degree_property_map("total"), mi=0.5, ma=22)
    layout = gt.sfdp_layout(g, vweight=deg, eweight=edge_weights)

    chemsys_names = [g.vp["entry"][v].chemsys for v in g.vertices()]

    edge_width = [1.4 if g.ep["cost"][e] != 0 else 0.1 for e in g.edges()]
    edge_width = g.new_edge_property("float", edge_width)

    color_func_v = _get_cmap_string(vertex_cmap_name, domain=sorted(chemsys_names))
    vertex_colors = [color_func_v(chemsys) for chemsys in chemsys_names]
    vertex_colors = g.new_vertex_property("vector<float>", vertex_colors)

    avg_cost = np.mean(costs)
    vmin = avg_cost - 0.8 * avg_cost
    vmax = avg_cost + 0.8 * avg_cost
    norm = matplotlib.colors.Normalize(vmin=vmin, vmax=vmax)
    edge_cmap = matplotlib.cm.get_cmap(edge_cmap_name)
    edge_cmap = matplotlib.cm.ScalarMappable(norm=norm, cmap=edge_cmap)

    edge_colors = [edge_cmap.to_rgba(cost, alpha=0.5) for cost in costs]
    edge_colors = g.new_edge_property("vector<float>", edge_colors)

    control = g.new_edge_property("vector<double>")
    for e in g.edges():
        d = np.sqrt(sum((layout[e.source()].a - layout[e.target()].a) ** 2)) / 20
        control[e] = [0.0, 0.0, 0.3, d, 0.7, d, 1.0, 0.0]

    return gt.graph_draw(
        g,
        pos=layout,
        vertex_fill_color=vertex_colors,
        edge_color=edge_colors,
        edge_pen_width=edge_width,
        vertex_size=deg,
        edge_control_points=control,
        output=output,
    )


def plot_network_on_graphistry(graph: gt.Graph):  # pragma: no cover
    """
    Plots a reaction network using Graphistry Hub (https://graphistry.org/). This uses
    optional dependencies networkx and pyintergraph to facilitate mapping to a format
    that can be easily rendered by graphistry.

    Args:
        graph: a graph-tool Graph object
    """

    warnings.warn(
        "This function may not work as expected, depending on your graphistry setup."
    )

    try:
        import graphistry
        import networkx as nx
        import pyintergraph
    except ImportError as e:
        raise ImportError(
            "Must install optional dependencies: pygraphistry, networkx, and"
            " pyintergraph!"
        ) from e

    nx_graph = pyintergraph.gt2nx(graph)
    mapping = {}

    for node in nx_graph.nodes(data=True):
        mapping[node[0]] = str(node[1]["entry"])
        nx_graph.nodes()[node[0]]["entry"] = str(nx_graph.nodes()[node[0]]["entry"])

    nx.relabel_nodes(nx_graph, mapping, copy=False)

    for edge in nx_graph.edges:
        nx_graph.edges[edge]["cost"] = float(nx_graph.edges[edge]["cost"])
        if nx_graph.edges[edge]["rxn"]:
            nx_graph.edges[edge]["rxn"] = str(nx_graph.edges[edge]["rxn"])
        else:
            nx_graph.edges[edge]["rxn"] = "None"

    return graphistry.bind(source="src", destination="dst", node="nodeid").plot(
        nx_graph, render=True
    )


def _get_cmap_string(palette, domain):
    """
    Utility function for getting a matplotlib colormap string for a given palette and
    domain.
    """
    domain_unique = np.unique(domain)
    hash_table = {key: i_str for i_str, key in enumerate(domain_unique)}
    mpl_cmap = matplotlib.cm.get_cmap(palette, lut=len(domain_unique))

    def cmap_out(X, **kwargs):
        return mpl_cmap(hash_table[X], **kwargs)

    return cmap_out



"""
Entry objects used in a Network. These network entry objects hold multiple entries and
can be used as data for a node in the graph.
"""
from enum import Enum, auto
from typing import List

from monty.json import MSONable
from pymatgen.entries import Entry


class NetworkEntryType(Enum):
    """Describes the Network Entry Type"""

    Precursors = auto()
    Reactants = auto()
    Products = auto()
    Target = auto()
    Dummy = auto()


class NetworkEntry(MSONable):
    """
    Helper class for describing combinations of ComputedEntry-like objects in context
    of a reaction network.
    """

    def __init__(self, entries: List[Entry], description: NetworkEntryType):
        """
        Args:
            entries: list of Entry-like objects
            description: Node type (e.g., Precursors, Target... see NetworkEntryType
                class)
        """
        self.entries = set(entries)
        self.elements = sorted(
            list({elem for entry in entries for elem in entry.composition.elements})
        )
        self.chemsys = "-".join([str(e) for e in self.elements])
        self.dim = len(self.chemsys)
        self.description = description

    def __repr__(self):
        formulas = [entry.composition.reduced_formula for entry in self.entries]
        formulas.sort()
        return f"{self.description.name}: {','.join(formulas)}"

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            if self.description == other.description:
                if self.chemsys == other.chemsys:
                    return self.entries == other.entries
        return False

    def __hash__(self):
        return hash((self.description, frozenset(self.entries)))


class DummyEntry(NetworkEntry):
    """
    A Dummy Entry that doesn't hold any info. This maybe useful for serving as an empty
    node to facilitate pathfinding to all nodes, etc.
    """

    def __init__(self):  # pylint: disable=super-init-not-called
        """Dummy node doesn't need any parameters"""

    @property
    def entries(self):
        """No entries in DummyEntry"""
        return []

    @property
    def chemsys(self):
        """No Chemsys to DummyEntry"""
        return ""

    @property
    def description(self):
        """DummyEntry is always of type Dummy"""
        return NetworkEntryType.Dummy

    def __repr__(self):
        return "Dummy Node"

    def __eq__(self, other):
        return self is other

    def __hash__(self):
        return hash("Dummy")



"""
This module implements added features to the ChemicalPotentialDiagram class from
pymatgen.
"""
from functools import cached_property
from typing import Dict, List, Optional, Tuple

import numpy as np
from pymatgen.analysis.chempot_diagram import ChemicalPotentialDiagram as ChempotDiagram
from pymatgen.analysis.phase_diagram import PDEntry, PhaseDiagram
from pymatgen.core.composition import Element
from scipy.spatial import HalfspaceIntersection, KDTree

from rxn_network.entries.entry_set import GibbsEntrySet


class ChemicalPotentialDiagram(ChempotDiagram):
    """
    This class is an extension of the ChemicalPotentialDiagram class from pymatgen.
    Several features have been added to the original class for the purpose of
    efficiently calculating the shortest distance between two chemical potential
    domains.

    For more information on this specific implementation of the algorithm,
    please cite/reference the paper below:

        Todd, P. K., McDermott, M. J., Rom, C. L., Corrao, A. A., Denney, J. J.,
        Dwaraknath, S. S.,  Khalifah, P. G., Persson, K. A., & Neilson, J. R. (2021).
        Selectivity in Yttrium Manganese Oxide Synthesis via Local Chemical Potentials
        in Hyperdimensional Phase Space. Journal of the American Chemical Society,
        143(37), 15185-15194. https://doi.org/10.1021/jacs.1c06229
    """

    def __init__(  # pylint: disable=super-init-not-called
        self,
        entries: List[PDEntry],
        limits: Optional[Dict[Element, float]] = None,
        default_min_limit: Optional[float] = -100.0,
    ):
        """
        Initialize a ChemicalPotentialDiagram object.

        Args:
            entries: List of PDEntry-like objects containing a composition and
                energy. Must contain elemental references and be suitable for typical
                phase diagram construction. Entries must be within a chemical system of
                with 2+ elements
            limits: Bounds of elemental chemical potentials (min, max), which are
                used to construct the border hyperplanes used in the
                HalfspaceIntersection algorithm; these constrain the space over which
                the domains are calculated and also determine the size of the plotted
                diagram. Any elemental limits not specified are covered in the
                default_min_limit argument
            default_min_limit (float): Default minimum chemical potential limit for
                unspecified elements within the "limits" argument. This results in
                default limits of (-100, 0)
        """
        self.entries = list(
            sorted(entries, key=lambda e: e.composition.reduced_formula)
        )
        self._entry_set = GibbsEntrySet(self.entries)
        self.limits = limits
        self.default_min_limit = default_min_limit
        self.elements = list(
            sorted({els for e in self.entries for els in e.composition.elements})
        )
        self.dim = len(self.elements)
        self._min_entries, self._el_refs = self._get_min_entries_and_el_refs(
            self.entries
        )
        self._entry_dict = {e.composition.reduced_formula: e for e in self._min_entries}
        self._border_hyperplanes = self._get_border_hyperplanes()
        (
            self._hyperplanes,
            self._hyperplane_entries,
        ) = self._get_hyperplanes_and_entries()

        if self.dim < 2:
            raise ValueError(
                "ChemicalPotentialDiagram currently requires phase "
                "diagrams with 2 or more elements!"
            )

        if len(self.el_refs) != self.dim:
            missing = set(self.elements).difference(self.el_refs.keys())
            raise ValueError(
                f"There are no entries for the terminal elements: {missing}"
            )
        self._hs_int = self._get_halfspace_intersection()

        num_hyperplanes = len(self._hyperplanes)
        num_border_hyperplanes = len(self._border_hyperplanes)

        self._border_hyperplane_indices = list(
            range(num_hyperplanes, num_hyperplanes + num_border_hyperplanes)
        )
        self._metastable_domains: Dict[str, List] = {}  # for caching

    def shortest_domain_distance(self, f1: str, f2: str, offset=0.0) -> float:
        """
        Args:
            f1: chemical formula (1)
            f2: chemical formula (2)

        Returns:
            Shortest distance between domain boundaries in the full
            (hyper)dimensional space, calculated using KDTree.
        """

        if f1 in self.domains:
            pts1 = self.domains[f1]
        else:
            pts1 = self._get_metastable_domain(f1)

        if f2 in self.domains:
            pts2 = self.domains[f2]
        else:
            pts2 = self._get_metastable_domain(f2)

        tree = KDTree(pts1)

        return min(tree.query(pts2)[0]) + offset

    def _get_halfspace_intersection(self):
        hs_hyperplanes = np.vstack([self._hyperplanes, self._border_hyperplanes])
        interior_point = np.min(self.lims, axis=1) + 1e-1
        return HalfspaceIntersection(hs_hyperplanes, interior_point)

    def _get_domains(self) -> Dict[str, np.ndarray]:
        """Returns a dictionary of domains as {formula: np.ndarray}"""
        domains: Dict[str, List] = {
            entry.composition.reduced_formula: [] for entry in self._hyperplane_entries
        }
        entries = self._hyperplane_entries

        for intersection, facet in zip(
            self.hs_int.intersections, self.hs_int.dual_facets
        ):
            for v in facet:
                if v not in self._border_hyperplane_indices:
                    this_entry = entries[v]
                    formula = this_entry.composition.reduced_formula
                    domains[formula].append(intersection)

        return {k: np.array(v) for k, v in domains.items() if v}

    def _get_hyperplanes_and_entries(self) -> Tuple[np.ndarray, List[PDEntry]]:
        """Returns both the array of hyperplanes, as well as a list of the minimum
        entries"""
        data = np.array([self._get_hyperplane(e) for e in self._min_entries])
        vec = [self.el_refs[el].energy_per_atom for el in self.elements] + [1]
        form_e = -np.dot(data, vec)

        inds = np.where(form_e < -PhaseDiagram.formation_energy_tol)[0].tolist()

        inds.extend([self._min_entries.index(el) for el in self.el_refs.values()])

        hyperplanes = data[inds]
        hyperplane_entries = [self._min_entries[i] for i in inds]

        return hyperplanes, hyperplane_entries

    def _get_hyperplane(self, entry):
        data = np.array(
            [entry.composition.get_atomic_fraction(el) for el in self.elements]
            + [-entry.energy_per_atom]
        )
        return data

    def get_offset(self, entry):
        """ """
        if (
            entry in self._min_entries
            and entry.composition.reduced_formula in self.domains
        ):
            offset = 0.0
        else:
            e_above_hull = self._entry_set.get_e_above_hull(entry)
            hyperplane = self._get_hyperplane(entry)
            offset = self._get_distance_between_parallel_hyperplanes(
                hyperplane[:-1], e_above_hull
            )

        return offset

    @staticmethod
    def _get_distance_between_parallel_hyperplanes(a, delta_b):
        """Returns the distance between two parallel hyperplanes"""
        return np.abs(delta_b) / np.linalg.norm(a)

    def _get_metastable_domain(self, formula):
        """Returns the metastable domain for a given formula"""
        if formula in self._metastable_domains:
            return self._metastable_domains[formula]

        orig_entry = self._entry_set.get_min_entry_by_formula(formula)
        new_entry = self._entry_set.get_stabilized_entry(orig_entry, tol=1e-1)
        self._entry_set.add(new_entry)
        cpd = ChemicalPotentialDiagram(self._entry_set, default_min_limit=-500)

        try:
            metastable_domain = cpd.domains[formula]
        except KeyError as exc:
            raise ValueError(
                f"Metastable domain for {formula} not found! Please investigate."
            ) from exc

        self._metastable_domains[formula] = metastable_domain
        self._entry_set.remove(new_entry)

        return metastable_domain

    @property
    def hs_int(self):
        """Returns the scipy HalfSpaceIntersection object"""
        return self._hs_int

    @cached_property
    def domains(self) -> dict[str, np.ndarray]:
        """Mapping of formulas to array of domain boundary points"""
        return self._get_domains()

    @property
    def metastable_domains(self) -> Dict[str, np.ndarray]:
        """
        Gets a dictionary of the chemical potential domains for metastable chemical
        formulas. This corresponds to the domains of the relevant phases if they were
        just barely stable
        """
        return {
            e.composition.reduced_formula: self._get_metastable_domain(
                e.composition.reduced_formula
            )
            for e in self._min_entries
            if e.composition.reduced_formula not in self.domains
        }



"""
Implements a class for conveniently and efficiently storing sets of ComputedReaction
objects which share entries.
"""
from collections import OrderedDict
from functools import lru_cache
from itertools import combinations, groupby
from typing import Any, Collection, Dict, Iterable, List, Optional, Set, Union

import numpy as np
from monty.json import MSONable
from pandas import DataFrame
from pymatgen.core.composition import Element
from pymatgen.entries.computed_entries import ComputedEntry

from rxn_network.core.composition import Composition
from rxn_network.core.cost_function import CostFunction
from rxn_network.reactions.computed import ComputedReaction
from rxn_network.reactions.open import OpenComputedReaction


class ReactionSet(MSONable):
    """
    A lightweight class for storing large sets of ComputedReaction objects.
    Automatically represents a set of reactions as an array of coefficients with
    a second array linking to a corresponding list of shared entries. This is useful for
    dumping large amounts of reaction data to a database.

    Note: this is not a true "set"; there is the option for filtering duplicates but it
        is not explicitly required.
    """

    def __init__(
        self,
        entries: List[ComputedEntry],
        indices: Union[np.ndarray, List[List[int]]],
        coeffs: Union[np.ndarray, List[List[float]]],
        open_elem: Optional[Union[str, Element]] = None,
        chempot: float = 0.0,
        all_data: Optional[List] = None,
    ):
        """
        Args:
            entries: List of ComputedEntry objects shared by reactions
            indices: Array indexing the entry list; gets entries used by each
                reaction object
            coeffs: Array of all reaction coefficients
            open_elem: Open element, e.g., "O"
            chempot: Chemical potential (mu) of open element in equation: Phi = G - mu*N
            all_data: Optional list of data for each reaction
        """
        self.entries = entries
        self.indices = indices
        self.coeffs = coeffs
        self.open_elem = open_elem
        self.chempot = chempot
        self.all_data = all_data if all_data else []

        self.mu_dict = None
        if open_elem:
            self.mu_dict = {Element(open_elem): chempot}  # type: ignore

    def get_rxns(
        self,
    ) -> Iterable[Union[ComputedReaction, OpenComputedReaction]]:
        """
        Returns list of ComputedReaction objects or OpenComputedReaction objects (when
        open element and chempot are specified) for the reaction set.
        """
        return self._get_rxns_by_indices(idxs=range(len(self.coeffs)))

    @classmethod
    def from_rxns(
        cls,
        rxns: Collection[Union[ComputedReaction, OpenComputedReaction]],
        entries: Optional[Collection[ComputedEntry]] = None,
        open_elem: Optional[Union[str, Element]] = None,
        chempot: float = 0.0,
        filter_duplicates: bool = False,
    ) -> "ReactionSet":
        """
        Initiate a ReactionSet object from a list of reactions. Including a list of
        unique entries saves some computation time.

        Args:
            rxns: List of ComputedReaction-like objects.
            entries: Optional list of ComputedEntry objects
            open_elem: Open element, e.g. "O2"
            chempot: Chemical potential (mu) of open element in equation: Phi = G - mu*N
        """

        if not entries:
            entries = cls._get_unique_entries(rxns)

        entries = sorted(list(set(entries)), key=lambda r: r.composition)

        all_entry_indices: Dict[str, ComputedEntry] = {}
        indices, coeffs, data = [], [], []

        for rxn in rxns:
            rxn_indices = []
            for e in rxn.entries:
                idx = all_entry_indices.get(e.composition.reduced_formula)
                if idx is None:
                    idx = entries.index(e)
                    all_entry_indices[e.composition.reduced_formula] = idx
                rxn_indices.append(idx)

            indices.append(rxn_indices)
            coeffs.append(list(rxn.coefficients))
            data.append(rxn.data)

        all_open_elems: Set[Element] = set()
        all_chempots: Set[float] = set()

        if (
            all(r.__class__.__name__ == "OpenComputedReaction" for r in rxns)
            and not open_elem
        ):
            for r in rxns:
                all_open_elems.update(r.chempots.keys())
                all_chempots.update(r.chempots.values())

            if len(all_chempots) == 1 and len(all_open_elems) == 1:
                chempot = all_chempots.pop()
                open_elem = all_open_elems.pop()

        rxn_set = cls(
            entries=entries,
            indices=indices,
            coeffs=coeffs,
            open_elem=open_elem,
            chempot=chempot,
            all_data=data,
        )

        if filter_duplicates:
            rxn_set = rxn_set.filter_duplicates()

        return rxn_set

    @lru_cache(maxsize=1)
    def to_dataframe(
        self,
        cost_function: CostFunction,
        target: Optional[Composition] = None,
        calculate_uncertainties=False,
        calculate_separable=False,
    ) -> DataFrame:
        """
        Make a dataframe of reactions from a ReactionSet object.

        Args:
            cost_function: Cost function to use for evaluating reaction costs
            target: Optional target composition (used to determine added elements)
            calculate_uncertainties: Whether to calculate uncertainties (dE column)
            calculate_separable: Whether to calculate if the reaction products are
                separable (see ComputedReaction.is_separable)

        Returns:
            Pandas DataFrame with columns:
                rxn: Reaction object
                energy: reaction energy in eV/atom
                dE (optional): uncertainty in reaction energy in eV/atom
                added_elems (optional): List of added elements
                separable (optional): whether reaction products are separable
                cost: Cost of reaction
                other: any other data associated with reaction

        """

        data: Dict[str, Any] = OrderedDict({k: [] for k in ["rxn", "energy"]})
        attrs = []

        calculate_e_above_hulls = False
        determine_theoretical = False

        for r in self.get_rxns():
            attrs = list(r.data.keys())  # get extra attributes from first reaction
            entry_data = r.entries[0].data
            if "e_above_hull" in entry_data:
                calculate_e_above_hulls = True
                data["max_e_hull_reactants"] = []
                data["max_e_hull_products"] = []
            if "icsd_ids" in entry_data or "theoretical" in entry_data:
                determine_theoretical = True
                data["num_theoretical_reactants"] = []
                data["num_theoretical_products"] = []

            break

        target = Composition(target) if target else None

        if "num_constraints" in attrs:
            attrs.remove("num_constraints")
        if calculate_uncertainties:
            data["dE"] = []
        if target:
            data["added_elems"] = []
            if calculate_separable:
                data["separable"] = []

        data["max_num_precursor_elems"] = []

        data.update({k: [] for k in attrs + ["cost"]})

        for rxn in self.get_rxns():
            data["rxn"].append(rxn)
            data["energy"].append(rxn.energy_per_atom)
            if calculate_uncertainties:
                data["dE"].append(rxn.energy_uncertainty_per_atom)
            if target:
                data["added_elems"].append(self._get_added_elems(rxn, target))
                if calculate_separable:
                    data["separable"].append(rxn.is_separable(target))

            if calculate_e_above_hulls:
                data["max_e_hull_reactants"].append(
                    max(e.data.get("e_above_hull", 0.0) for e in rxn.reactant_entries)
                )
                data["max_e_hull_products"].append(
                    max(e.data.get("e_above_hull", 0.0) for e in rxn.product_entries)
                )
            if determine_theoretical:
                data["num_theoretical_reactants"].append(
                    sum(bool(not e.is_experimental) for e in rxn.reactant_entries)
                )
                data["num_theoretical_products"].append(
                    sum(bool(not e.is_experimental) for e in rxn.product_entries)
                )

            data["max_num_precursor_elems"].append(
                max(len(precursor.elements) for precursor in rxn.reactants)
            )

            for attr in attrs:
                data[attr].append(rxn.data.get(attr))

            data["cost"].append(cost_function.evaluate(rxn))

        df = DataFrame(data).sort_values("cost").reset_index(drop=True)
        return df

    def calculate_costs(
        self,
        cf: CostFunction,
    ) -> List[float]:
        """
        Evaluate a cost function on an acquired set of reactions.

        Args:
            cf: CostFunction object, e.g. Softplus()
        """
        return [cf.evaluate(rxn) for rxn in self.get_rxns()]

    def add_rxns(self, rxns):
        """
        Return a new ReactionSet with the reactions added.

        Warning: all new reactions must only have entires contained in the entries of
        the current reaction set.
        """
        new_indices, new_coeffs, new_data = [], [], []
        for rxn in rxns:
            new_indices.append([self.entries.index(e) for e in rxn.entries])
            new_coeffs.append(list(rxn.coefficients))
            new_data.append(rxn.data)

        return ReactionSet(
            self.entries,
            self.indices + new_indices,
            self.coeffs + new_coeffs,
            self.open_elem,
            self.chempot,
            self.all_data + new_data,
        )

    def add_rxn_set(self, rxn_set):
        """Adds a new reaction set to current reaction set.

        Warning: new reaction set must have the same entries as the current reaction
        set.

        """
        if self.entries != rxn_set.entries:
            raise ValueError(
                "Reaction sets must have identical entries property to add."
            )
        open_elem = self.open_elem
        chempot = self.chempot

        indices = self.indices + rxn_set.indices
        coeffs = self.coeffs + rxn_set.coeffs
        all_data = self.all_data + rxn_set.all_data

        return ReactionSet(self.entries, indices, coeffs, open_elem, chempot, all_data)

    def get_rxns_by_reactants(self, reactants: List[str]):
        """
        Return a list of reactions with the given reactants.
        """
        idxs = []
        reactants = [Composition(r).reduced_formula for r in reactants]

        reactant_indices = {
            idx
            for idx, e in enumerate(self.entries)
            if e.composition.reduced_formula in reactants
        }

        if not reactant_indices:
            return []

        for idx, (coeffs, indices) in enumerate(zip(self.coeffs, self.indices)):
            r_indices = {i for c, i in zip(coeffs, indices) if c < 1e-12}
            if r_indices.issubset(reactant_indices):
                idxs.append(idx)

        return self._get_rxns_by_indices(idxs)

    def get_rxns_by_product(self, product: str):
        """
        Return a list of reactions which contain the given product formula.
        """
        idxs = []
        product = Composition(product).reduced_formula

        product_index = None
        for idx, e in enumerate(self.entries):
            if e.composition.reduced_formula == product:
                product_index = idx
                break

        if not product_index:
            return []

        for idx, (coeffs, indices) in enumerate(zip(self.coeffs, self.indices)):
            p_indices = {i for c, i in zip(coeffs, indices) if c > -1e-12}
            if product_index in p_indices:
                idxs.append(idx)

        return self._get_rxns_by_indices(idxs)

    def filter_duplicates(self):
        """
        Return a new ReactionSet object with duplicate reactions removed
        """
        indices_to_remove = set()
        if len(self.coeffs) == 0:
            return self

        # groupby only works with pre-sorted arrays
        sorted_coeffs, sorted_idxs, sorted_indices = zip(
            *list(
                sorted(
                    zip(self.coeffs, range(len(self.indices)), self.indices),
                    key=lambda x: sorted(x[2]),
                )
            )
        )

        for i, group in groupby(
            zip(sorted_coeffs, sorted_idxs, sorted_indices),
            key=lambda i: sorted(i[2]),
        ):
            coeffs_group, idx_group, indices_group = zip(*group)
            if len(idx_group) > 1:
                for (_, coeffs1, indices1), (
                    idx2,
                    coeffs2,
                    indices2,
                ) in combinations(zip(idx_group, coeffs_group, indices_group), 2):
                    if idx2 in indices_to_remove:
                        continue

                    coeffs2_sorted = [coeffs2[indices2.index(i)] for i in indices1]

                    ratios = np.array(coeffs1) / np.array(coeffs2_sorted)
                    if (
                        ratios <= 1e-8
                    ).any():  # do not remove any reaction with negative ratio
                        continue

                    if np.isclose(ratios[0], ratios).all():
                        indices_to_remove.add(idx2)

        new_indices = []
        new_coeffs = []
        new_all_data = []

        for idx in set(range(len(self))) - indices_to_remove:
            new_indices.append(self.indices[idx])
            new_coeffs.append(self.coeffs[idx])
            new_all_data.append(self.all_data[idx])

        return ReactionSet(
            self.entries,
            new_indices,
            new_coeffs,
            self.open_elem,
            self.chempot,
            new_all_data,
        )

    def _get_rxns_by_indices(
        self, idxs: Union[List[int], range]
    ) -> Iterable[Union[ComputedReaction, OpenComputedReaction]]:
        """
        Return a list of reactions with the given indices.
        """
        if idxs == range(len(self.coeffs)):
            indices_slice = self.indices
            coeffs_slice = self.coeffs
            data_slice = self.all_data
        else:
            indices_slice = []
            coeffs_slice = []
            data_slice = []

            for idx in idxs:
                indices_slice.append(self.indices[idx])
                coeffs_slice.append(self.coeffs[idx])
                data_slice.append(self.all_data[idx])

        for indices, coeffs, data in zip(indices_slice, coeffs_slice, data_slice):
            entries = [self.entries[i] for i in indices]
            if self.mu_dict:
                rxn = OpenComputedReaction(
                    entries=entries,
                    coefficients=coeffs,
                    data=data,
                    chempots=self.mu_dict,
                )
            else:
                rxn = ComputedReaction(entries=entries, coefficients=coeffs, data=data)

            yield rxn

    @staticmethod
    def _get_added_elems(
        rxn: Union[ComputedReaction, OpenComputedReaction], target: Composition
    ) -> str:
        """
        Get list of added elements for a reaction.

        Args:
            rxn: Reaction object
            target: target composition (used to determine added elements)

        """
        target = Composition(target)
        chemsys_prop = (
            "total_chemical_system"
            if rxn.__class__.__name__ == "OpenComputedReaction"
            else "chemical_system"
        )

        added_elems = set(getattr(rxn, chemsys_prop).split("-")) - set(
            target.chemical_system.split("-")
        )
        added_elems_str = "-".join(sorted(list(added_elems)))

        return added_elems_str

    @staticmethod
    def _get_unique_entries(rxns: Collection[ComputedReaction]) -> Set[ComputedEntry]:
        """
        Return only unique entries from reactions
        """
        entries = set()
        for r in rxns:
            entries.update(r.entries)
        return entries

    def __iter__(self):
        """
        Iterate over the reactions in the set.
        """
        return iter(self.get_rxns())

    def __len__(self):
        """
        Return length of reactions stored in the set.
        """
        return len(self.coeffs)



"""
Implements a reaction pathway solver class which efficiently solves mass balance
equations using matrix operations.
"""

from copy import deepcopy
from itertools import combinations
from math import comb
from typing import Union

import numpy as np
import ray
from numba import njit, prange
from pymatgen.core.composition import Element
from tqdm import tqdm

from rxn_network.core.composition import Composition
from rxn_network.core.cost_function import CostFunction
from rxn_network.core.solver import Solver
from rxn_network.entries.entry_set import GibbsEntrySet
from rxn_network.enumerators.basic import BasicEnumerator, BasicOpenEnumerator
from rxn_network.enumerators.minimize import (
    MinimizeGibbsEnumerator,
    MinimizeGrandPotentialEnumerator,
)
from rxn_network.pathways.balanced import BalancedPathway
from rxn_network.pathways.pathway_set import PathwaySet
from rxn_network.reactions.computed import ComputedReaction
from rxn_network.reactions.open import OpenComputedReaction
from rxn_network.reactions.reaction_set import ReactionSet
from rxn_network.utils.funcs import grouper
from rxn_network.utils.ray import initialize_ray, to_iterator


class PathwaySolver(Solver):
    """
    Solver that implements an efficient method (using numba) for finding balanced
    reaction pathways from a list of graph-derived reaction pathways (i.e. a list of
    lists of reactions)
    """

    def __init__(
        self,
        pathways: PathwaySet,
        entries: GibbsEntrySet,
        cost_function: CostFunction,
        open_elem: str = None,
        chempot: float = None,
        chunk_size=100000,
        batch_size=None,
    ):
        """
        Args:
            pathways: List of reaction pathways derived from the network.
            entries: GibbsEntrySet containing all entries in the network.
            cost_function: CostFunction object to use for the solver.
            open_elem: Element to use for pathways with an open element.
            chempot: Chemical potential to use for pathways with an open element.
        """
        super().__init__(pathways=deepcopy(pathways))
        self._entries = entries
        self.cost_function = cost_function
        self.open_elem = Element(open_elem) if open_elem else None
        self.chempot = chempot
        self.chunk_size = chunk_size
        self.batch_size = batch_size

    def solve(
        self,
        net_rxn: Union[ComputedReaction, OpenComputedReaction],
        max_num_combos: int = 4,
        find_intermediate_rxns: bool = True,
        intermediate_rxn_energy_cutoff: float = 0.0,
        use_basic_enumerator: bool = True,
        use_minimize_enumerator: bool = False,
        filter_interdependent: bool = True,
    ) -> PathwaySet:
        """

        Args:
            net_rxn: The reaction representing the total reaction from precursors to
                final targets.
            max_num_combos: The maximum allowable size of the balanced reaction pathway.
                At values <=5, the solver will start to take a significant amount of
                time to run.
            find_intermediate_rxns: Whether to find intermediate reactions; crucial for
                finding pathways where intermediates react together, as these reactions
                may not occur in the graph-derived pathways. Defaults to True.
            intermediate_rxn_energy_cutoff: An energy cutoff by which to filter down
                intermediate reactions. This can be useful when there are a large number
                of possible intermediates. < 0 means allow only exergonic reactions.
            use_basic_enumerator: Whether to use the BasicEnumerator to find
                intermediate reactions. Defaults to True.
            use_minimize_enumerator: Whether to use the MinimizeGibbsEnumerator to find
                intermediate reactions. Defaults to False.
            filter_interdependent: Whether or not to filter out pathways where reaction
                steps are interdependent. Defaults to True.

        Returns:
            A list of BalancedPathway objects.
        """

        if not net_rxn.balanced:
            raise ValueError(
                "Net reaction must be balanceable to find all reaction pathways."
            )

        initialize_ray()

        entries_copy = deepcopy(self.entries)
        entries = entries_copy.entries_list
        num_entries = len(entries)

        reactions = deepcopy(self.reactions)
        costs = deepcopy(self.costs)

        precursors = deepcopy(net_rxn.reactant_entries)
        targets = deepcopy(net_rxn.product_entries)

        self.logger.info(f"Net reaction: {net_rxn} \n")

        if find_intermediate_rxns:
            self.logger.info("Identifying reactions between intermediates...")

            intermediate_rxns = self._find_intermediate_rxns(
                targets,
                intermediate_rxn_energy_cutoff,
                use_basic_enumerator,
                use_minimize_enumerator,
            )
            intermediate_costs = [
                self.cost_function.evaluate(r) for r in intermediate_rxns.get_rxns()
            ]
            for r, c in zip(intermediate_rxns, intermediate_costs):
                if r not in reactions:
                    reactions.append(r)
                    costs.append(c)

        net_rxn_vector = net_rxn.get_entry_idx_vector(num_entries)

        if net_rxn in reactions:
            reactions.remove(net_rxn)

        reaction_set = ray.put(ReactionSet.from_rxns(reactions))
        entries = ray.put(entries)
        costs = ray.put(costs)
        num_entries = ray.put(num_entries)
        net_rxn_vector = ray.put(net_rxn_vector)
        open_elem = ray.put(self.open_elem)
        chempot = ray.put(self.chempot)

        num_rxns = len(reactions)
        batch_size = self.batch_size or ray.cluster_resources()["CPU"] - 1

        num_combos = sum(comb(num_rxns, k) for k in range(1, max_num_combos + 1))
        num_batches = int((num_combos // self.chunk_size + 1) // batch_size + 1)

        paths = []
        paths_refs = []
        batch_count = 1
        for n in range(1, max_num_combos + 1):
            for group in grouper(combinations(range(num_rxns), n), self.chunk_size):
                paths_refs.append(
                    _get_balanced_paths_ray.remote(
                        group,
                        reaction_set,
                        costs,
                        entries,
                        num_entries,
                        net_rxn_vector,
                        open_elem,
                        chempot,
                    )
                )
                if len(paths_refs) >= batch_size:
                    for paths_ref in tqdm(
                        to_iterator(paths_refs),
                        total=len(paths_refs),
                        desc=(
                            f"{self.__class__.__name__} (Batch"
                            f" {batch_count}/{num_batches})"
                        ),
                    ):
                        paths.extend(paths_ref)

                    batch_count += 1

                    paths_refs = []

        for paths_ref in tqdm(
            to_iterator(paths_refs),
            total=len(paths_refs),
            desc=f"{self.__class__.__name__} (Batch {batch_count}/{num_batches})",
        ):
            paths.extend(paths_ref)

        filtered_paths = []
        if filter_interdependent:
            precursor_comps = [p.composition for p in precursors]
            for p in paths:
                interdependent = p.contains_interdependent_rxns(precursor_comps)
                if not interdependent:
                    filtered_paths.append(p)
        else:
            filtered_paths = paths

        filtered_paths = sorted(list(set(filtered_paths)), key=lambda p: p.average_cost)

        return PathwaySet.from_paths(filtered_paths)

    def _find_intermediate_rxns(
        self,
        targets,
        energy_cutoff,
        use_basic_enumerator,
        use_minimize_enumerator,
    ):
        """
        Method for finding intermediate reactions using enumerators and
        specified settings.
        """

        intermediates = {e for rxn in self.reactions for e in rxn.entries}
        intermediates = GibbsEntrySet(
            list(intermediates) + targets,
        )
        target_formulas = [e.composition.reduced_formula for e in targets]
        ref_elems = {e for e in self.entries if e.is_element}

        intermediates = intermediates | ref_elems

        rxn_set = ReactionSet(
            intermediates.entries_list,
            [],
            [],
            open_elem=self.open_elem,
            chempot=self.chempot,
            all_data=[],
        )

        if use_basic_enumerator:
            be = BasicEnumerator(targets=target_formulas, calculate_e_above_hulls=False)
            rxn_set = rxn_set.add_rxn_set(be.enumerate(intermediates))

            if self.open_elem:
                boe = BasicOpenEnumerator(
                    open_phases=[Composition(str(self.open_elem)).reduced_formula],
                    targets=target_formulas,
                    calculate_e_above_hulls=False,
                )

                rxn_set = rxn_set.add_rxn_set(boe.enumerate(intermediates))

        if use_minimize_enumerator:
            mge = MinimizeGibbsEnumerator(
                targets=target_formulas, calculate_e_above_hulls=False
            )
            rxn_set = rxn_set.add_rxn_set(mge.enumerate(intermediates))

            if self.open_elem:
                mgpe = MinimizeGrandPotentialEnumerator(
                    open_elem=self.open_elem,
                    mu=self.chempot,
                    targets=target_formulas,
                )
                rxn_set.add_rxn_set(mgpe.enumerate(intermediates))

        rxns = list(filter(lambda x: x.energy_per_atom < energy_cutoff, rxn_set))
        rxns = [r for r in rxns if all(e in intermediates for e in r.entries)]
        num_rxns = len(rxns)
        rxns = ReactionSet.from_rxns(rxns, filter_duplicates=True)

        self.logger.info(f"Found {num_rxns} intermediate reactions! \n")

        return rxns

    @property
    def entries(self) -> GibbsEntrySet:
        """Entry set used in solver"""
        return self._entries


@njit(parallel=True, fastmath=True)
def _balance_path_arrays(
    comp_matrices: np.ndarray,
    net_coeffs: np.ndarray,
    tol: float = 1e-6,
):
    """
    Fast solution for reaction multiplicities via mass balance stochiometric
    constraints. Parallelized using Numba. Can be applied to large batches (100K-1M
    sets of reactions at a time.)

    Args:
        comp_matrices: Array containing stoichiometric coefficients of all
            compositions in all reactions, for each trial combination.
        net_coeffs: Array containing stoichiometric coefficients of net reaction.
        tol: numerical tolerance for determining if a multiplicity is zero
            (reaction was removed).
    """
    shape = comp_matrices.shape
    net_coeff_filter = np.argwhere(net_coeffs != 0).flatten()
    len_net_coeff_filter = len(net_coeff_filter)
    all_multiplicities = np.zeros((shape[0], shape[1]), np.float64)
    indices = np.full(shape[0], False)

    for i in prange(shape[0]):  # pylint: disable=not-an-iterable
        correct = True
        for j in range(len_net_coeff_filter):
            idx = net_coeff_filter[j]
            if not comp_matrices[i][:, idx].any():
                correct = False
                break
        if not correct:
            continue

        comp_pinv = np.linalg.pinv(comp_matrices[i]).T
        multiplicities = comp_pinv @ net_coeffs
        solved_coeffs = comp_matrices[i].T @ multiplicities

        if (multiplicities < tol).any():
            continue

        if not (
            np.abs(solved_coeffs - net_coeffs) <= (1e-08 + 1e-05 * np.abs(net_coeffs))
        ).all():
            continue

        all_multiplicities[i] = multiplicities
        indices[i] = True

    filtered_indices = np.argwhere(indices != 0).flatten()
    length = filtered_indices.shape[0]
    filtered_comp_matrices = np.empty((length, shape[1], shape[2]), np.float64)
    filtered_multiplicities = np.empty((length, shape[1]), np.float64)

    for i in range(length):
        idx = filtered_indices[i]
        filtered_comp_matrices[i] = comp_matrices[idx]
        filtered_multiplicities[i] = all_multiplicities[idx]

    return filtered_comp_matrices, filtered_multiplicities


def _create_comp_matrices(combos, rxns, num_entries):
    """Create array of stoichiometric coefficients for each reaction."""
    comp_matrices = np.stack(
        [
            np.vstack([rxns[r].get_entry_idx_vector(num_entries) for r in combo])
            for combo in combos
            if combo
        ]
    )
    return comp_matrices


@ray.remote
def _get_balanced_paths_ray(
    combos,
    reaction_set,
    costs,
    entries,
    num_entries,
    net_rxn_vector,
    open_elem,
    chempot,
):
    reactions = list(reaction_set.get_rxns())
    comp_matrices = _create_comp_matrices(combos, reactions, num_entries)

    paths = []

    c_mats, m_mats = _balance_path_arrays(comp_matrices, net_rxn_vector)

    for c_mat, m_mat in zip(c_mats, m_mats):
        path_rxns = []
        path_costs = []

        for rxn_mat in c_mat:
            ents, coeffs = zip(
                *[
                    (entries[idx], c)
                    for idx, c in enumerate(rxn_mat)
                    if not np.isclose(c, 0.0)
                ]
            )

            if open_elem is not None:
                rxn = OpenComputedReaction(
                    entries=ents,
                    coefficients=coeffs,
                    chempots={open_elem: chempot},
                )

            else:
                rxn = ComputedReaction(entries=ents, coefficients=coeffs)

            try:
                path_rxns.append(rxn)
                path_costs.append(costs[reactions.index(rxn)])
            except Exception as e:
                print(e)
                continue

        p = BalancedPathway(path_rxns, m_mat.flatten(), path_costs, balanced=True)
        paths.append(p)

    return paths



"""
This script generates the data/freed/compounds.json.gz file by using the s4 package as an interface.
"""

import gzip
import json
from collections import OrderedDict

import numpy as np
from s4.thermo.exp.freed import ExpThermoDatabase

if __name__ == "__main__":
    xp = ExpThermoDatabase()
    temps = np.arange(300, 2100, 100)
    data = {}  # type: ignore

    # Acquire data
    for c, phases in xp.compositions.items():
        if c.is_element:
            continue

        formula, factor = c.get_integer_formula_and_factor()

        if not data.get(formula):
            data[formula] = dict()

        for t in temps:
            energies = []
            for p in phases:
                if t > p.tmax:
                    continue
                energies.append(p.dgf(t))

            if not energies:
                continue

            final_energy = min(energies)
            g = final_energy / factor

            if data[formula].get(t):
                data[formula][t].append(g)
            else:
                data[formula][t] = [g]

    # Clean the data
    cleaned_data = {}
    for f, energy_dict in data.items():
        cleaned_data[f] = {str(t): min(e) for t, e in energy_dict.items()}

    # Write to file
    cleaned_data = OrderedDict(sorted(cleaned_data.items()))
    with gzip.open("compounds.json.gz", "wt", encoding="ascii") as zipfile:
        json.dump(cleaned_data, zipfile)



"""Core definition for various task and synthesis recipe documents"""
from datetime import datetime
from typing import List, Optional

from pydantic import BaseModel, Field
from pymatgen.core.composition import Element

from rxn_network.core.enumerator import Enumerator
from rxn_network.core.network import Network
from rxn_network.core.solver import Solver
from rxn_network.entries.entry_set import GibbsEntrySet
from rxn_network.pathways.pathway_set import PathwaySet
from rxn_network.reactions.computed import ComputedReaction
from rxn_network.reactions.reaction_set import ReactionSet
from rxn_network.utils.funcs import datetime_str


class EntrySetDocument(BaseModel):

    task_label: str = Field(None, description="The name of the task.")
    last_updated: datetime = Field(
        default_factory=datetime_str,
        description="Timestamp of when the document was last updated.",
    )
    entries: GibbsEntrySet = Field(description="The entry set")
    e_above_hull: float = Field(None, description="The e_above_hull cutoff")
    include_polymorphs: bool = Field(False, description="Whether to include polymorphs")
    formulas_to_include: Optional[List[str]] = Field(
        None, description="The formulas to include"
    )


class EnumeratorTaskDocument(BaseModel):
    """
    A single task object from the reaction enumerator workflow.
    """

    task_label: str = Field(None, description="The name of the task.")
    last_updated: datetime = Field(
        default_factory=datetime_str,
        description="Timestamp of when the document was last updated.",
    )
    rxns: ReactionSet = Field(description="The reaction set.")
    targets: List[str] = Field(
        None, description="The target formulas used in the enumerator(s)."
    )
    elements: List[Element] = Field(
        None, description="The elements of the total chemical system"
    )
    chemsys: str = Field(
        None, description="The total chemical system string (e.g., Fe-Li-O)."
    )
    added_elements: List[Element] = Field(
        None, description="The elements added beyond the elements of the target(s)."
    )
    added_chemsys: str = Field(
        None, description="The chemical system of the added elements"
    )
    enumerators: List[Enumerator] = Field(
        None,
        description="A list of the enumerator objects used to calculate the reactions.",
    )


class NetworkTaskDocument(BaseModel):
    """
    The calculation output from the NetworkMaker workflow. Contains the ReactionNetwork
    object and a link to the file where the graph-tool Graph object is stored.
    Optional: includes unbalanced paths found from pathfinding.
    """

    task_label: str = Field(None, description="The name of the task.")
    last_updated: datetime = Field(
        default_factory=datetime_str,
        description="Timestamp of when the document was last updated.",
    )
    network: Network = Field(description="The reaction network")
    graph_fn: str = Field(description="The file path of the network graph")
    paths: PathwaySet = Field(None, description="The (simple) reaction pathways")
    k: int = Field(None, description="The number of paths solved for")
    precursors: List[str] = Field(None, description="The precursor compositions")
    targets: List[str] = Field(None, description="The target compositions")


class PathwaySolverTaskDocument(BaseModel):
    """
    The calculation output from the PathwaySolverMaker workflow. Contains the pathway
    solver object and its calculated balanced pathways.
    """

    task_label: str = Field(None, description="The name of the task.")
    last_updated: datetime = Field(
        default_factory=datetime_str,
        description="Timestamp of when the document was last updated.",
    )
    solver: Solver = Field(description="The pathway solver used to calculate pathways")
    balanced_paths: PathwaySet = Field(description="The balanced reaction pathways")
    precursors: list[str] = Field(description="The precursor compositions")
    targets: list[str] = Field(description="The target compositions")
    net_rxn: ComputedReaction = Field(
        description="The net reaction used for pathway solving"
    )
    max_num_combos: int = Field(
        description=(
            "The maximum number of combinations to consider in the pathway solver"
        )
    )
    find_intermediate_rxns: bool = Field(
        description="Whether to find reactions from intermediate compositions"
    )
    intermediate_rxn_energy_cutoff: float = Field(
        description="The mximum energy cutoff for filtering intermediate reactions"
    )
    use_basic_enumerator: bool = Field(
        description="Whether to use the basic enumerators in path solving"
    )
    use_minimize_enumerator: bool = Field(
        description="Whether to use the minimize enumerators in path solving"
    )
    filter_interdependent: bool = Field(
        description="Whether to filter out interdependent pathways"
    )



"""Core jobs for reaction-network creation and analysis."""

from dataclasses import dataclass, field
from pathlib import Path
from typing import Iterable, List, Optional

from jobflow import SETTINGS, Maker, job
from pymatgen.core.composition import Element

from rxn_network.core.composition import Composition
from rxn_network.core.cost_function import CostFunction
from rxn_network.costs.softplus import Softplus
from rxn_network.entries.utils import get_all_entries_in_chemsys, process_entries
from rxn_network.enumerators.utils import get_computed_rxn
from rxn_network.jobs.schema import (
    EntrySetDocument,
    EnumeratorTaskDocument,
    NetworkTaskDocument,
    PathwaySolverTaskDocument,
)
from rxn_network.jobs.utils import get_added_elem_data, run_enumerators
from rxn_network.network.network import ReactionNetwork
from rxn_network.pathways.solver import PathwaySolver
from rxn_network.reactions.basic import BasicReaction
from rxn_network.reactions.reaction_set import ReactionSet
from rxn_network.utils.funcs import get_logger

logger = get_logger(__name__)


@dataclass
class GetEntrySetMaker(Maker):
    """
    Maker to create job for acquiring and processing entries to be used in reaction
    enumeration or network building.

    Args (A list of argumnets provided to the dataclass)
        name: Name of the job.
        entry_db_name: Name of the entry database store to use. If none is available,
            will automatically use MPRester to acquire entries.
        temperature: Temperature to use for computing thermodynamic properties.
        include_nist_data: Whether to include NIST data in the entry set.
        include_barin_data: Whether to include Barin data in the entry set.
        include_freed_data: Whether to include FREED data in the entry set.
        e_above_hull: Energy above hull to use for filtering entries.
        include_polymorphs: Whether to include polymorphs in the entry set.
        formulas_to_include: List of formulas to include in the entry set.
        calculate_e_above_hulls: Whether to calculate e_above_hulls for all entries
            in the entry set.
        MP_API_KEY: API key for Materials Project. Note: if not provided, MPRester will
            automatically look for an environment variable.
    """

    name: str = "get_and_process_entries"
    entry_db_name: str = "entries_db"
    temperature: int = 300
    include_nist_data: bool = True
    include_barin_data: bool = False
    include_freed_data: bool = False
    e_above_hull: float = 0.0
    include_polymorphs: bool = False
    formulas_to_include: list = field(default_factory=list)
    calculate_e_above_hulls: bool = True
    MP_API_KEY: Optional[str] = None
    property_data: Optional[List[str]] = None

    @job(entries="entries", output_schema=EntrySetDocument)
    def make(self, chemsys):
        entry_db = SETTINGS.JOB_STORE.additional_stores.get(self.entry_db_name)

        property_data = self.property_data
        if property_data is None:
            property_data = ["theoretical"]
        elif "theoretical" not in property_data:
            property_data.append("theoretical")

        if entry_db:
            entries = get_all_entries_in_chemsys(
                entry_db,
                chemsys,
                inc_structure=True,
                compatible_only=True,
                property_data=property_data,
                use_premade_entries=False,
            )
        else:
            from mp_api.client import MPRester

            if self.MP_API_KEY:
                with MPRester(self.MP_API_KEY) as mpr:
                    entries = mpr.get_entries_in_chemsys(elements=chemsys)
            else:
                with MPRester() as mpr:  # let MPRester look for env variable
                    entries = mpr.get_entries_in_chemsys(elements=chemsys)

        entries = process_entries(
            entries,
            temperature=self.temperature,
            include_nist_data=self.include_nist_data,
            include_barin_data=self.include_barin_data,
            include_freed_data=self.include_freed_data,
            e_above_hull=self.e_above_hull,
            include_polymorphs=self.include_polymorphs,
            formulas_to_include=self.formulas_to_include,
            calculate_e_above_hulls=self.calculate_e_above_hulls,
        )

        doc = EntrySetDocument(
            entries=entries,
            e_above_hull=self.e_above_hull,
            include_polymorphs=self.include_polymorphs,
            formulas_to_include=self.formulas_to_include,
        )
        doc.task_label = self.name

        return doc


@dataclass
class ReactionEnumerationMaker(Maker):
    """
    Maker to create job for enumerating reactions from a set of entries.
    """

    name: str = "enumerate_reactions"

    @job(rxns="rxns", output_schema=EnumeratorTaskDocument)
    def make(self, enumerators, entries):
        data = {}

        logger.info("Running enumerators...")
        data["rxns"] = run_enumerators(enumerators, entries)
        data.update(self._get_metadata(enumerators, entries))

        logger.info("Building task document...")
        enumerator_task = EnumeratorTaskDocument(**data)
        enumerator_task.task_label = self.name
        return enumerator_task

    def _get_metadata(self, enumerators, entries):
        chemsys = "-".join(entries.chemsys)
        targets = {
            target for enumerator in enumerators for target in enumerator.targets
        }

        added_elements = None
        added_chemsys = None

        if targets:
            added_elements, added_chemsys = get_added_elem_data(entries, targets)

        metadata = {
            "elements": [Element(e) for e in chemsys.split("-")],
            "chemsys": chemsys,
            "enumerators": [e.as_dict() for e in enumerators],
            "targets": list(sorted(targets)),
            "added_elements": added_elements,
            "added_chemsys": added_chemsys,
        }
        return metadata


@dataclass
class NetworkMaker(Maker):
    """
    Maker for generating reaction networks and performing pathfinding from a set of
    previously enumerated reactions.
    """

    name: str = "build_and_analyze_network"
    cost_function: CostFunction = field(default_factory=Softplus)
    precursors: Optional[List[str]] = None
    targets: Optional[List[str]] = None
    calculate_pathways: Optional[int] = 10
    open_elem: Optional[Element] = None
    chempot: float = 0.0
    graph_fn: Optional[str] = None

    @job(network="network", output_schema=NetworkTaskDocument)
    def make(
        self,
        rxn_sets: Iterable[ReactionSet],
    ):
        all_rxns = ReactionSet.from_rxns(
            [rxn for rxn_set in rxn_sets for rxn in rxn_set.get_rxns()],
            open_elem=self.open_elem,
            chempot=self.chempot,
        )
        all_rxns = all_rxns.filter_duplicates()

        rn = ReactionNetwork(all_rxns, cost_function=self.cost_function)
        rn.build()

        paths = None
        if self.precursors:
            rn.set_precursors(self.precursors)
        if self.targets:
            rn.set_target(self.targets[0])
        if self.calculate_pathways and self.targets:
            paths = rn.find_pathways(self.targets, k=self.calculate_pathways)

        graph_fn = self.graph_fn or "network.gt.gz"
        graph_fn = str(Path(graph_fn).absolute())
        rn.graph.save(graph_fn)

        data = {
            "network": rn,
            "graph_fn": graph_fn,
            "paths": paths,
            "k": self.calculate_pathways,
            "precursors": self.precursors,
            "targets": self.targets,
        }
        doc = NetworkTaskDocument(**data)
        doc.task_label = self.name
        return doc


@dataclass
class PathwaySolverMaker(Maker):
    """
    Maker for solving balanced reaction pathways from a set of (unbalanced) pathways
    returned as part of reaction network analysis.
    """

    name: str = "solve pathways"
    cost_function: CostFunction = field(default_factory=Softplus)
    precursors: list[str] = field(default_factory=list)
    targets: list[str] = field(default_factory=list)
    open_elem: Optional[Element] = None
    chempot: Optional[float] = None
    max_num_combos: int = 4
    find_intermediate_rxns: bool = True
    intermediate_rxn_energy_cutoff: float = 0.0
    use_basic_enumerator: bool = True
    use_minimize_enumerator: bool = False
    filter_interdependent: bool = True

    def __post_init__(self):
        net_rxn = BasicReaction.balance(
            [Composition(r) for r in self.precursors],
            [Composition(p) for p in self.targets],
        )
        if not net_rxn.balanced:
            raise ValueError(
                "Can not balance pathways with specified precursors/targets. Please"
                " make sure a balanced net reaction can be written from the provided"
                " reactant and product formulas!"
            )
        self.net_rxn = net_rxn

    @job(paths="paths", output_schema=PathwaySolverTaskDocument)
    def make(self, pathways, entries):
        chempots = None
        if self.open_elem:
            chempots = {Element(self.open_elem): self.chempot}

        ps = PathwaySolver(
            pathways=pathways,
            entries=entries,
            cost_function=self.cost_function,
            open_elem=self.open_elem,
            chempot=self.chempot,
        )

        net_rxn = get_computed_rxn(self.net_rxn, entries, chempots)

        balanced_paths = ps.solve(
            net_rxn=net_rxn,
            max_num_combos=self.max_num_combos,
            find_intermediate_rxns=self.find_intermediate_rxns,
            intermediate_rxn_energy_cutoff=self.intermediate_rxn_energy_cutoff,
            use_basic_enumerator=self.use_basic_enumerator,
            use_minimize_enumerator=self.use_minimize_enumerator,
            filter_interdependent=self.filter_interdependent,
        )
        data = {
            "solver": ps,
            "balanced_paths": balanced_paths,
            "precursors": self.precursors,
            "targets": self.targets,
            "net_rxn": net_rxn,
            "max_num_combos": self.max_num_combos,
            "find_intermediate_rxns": self.find_intermediate_rxns,
            "intermediate_rxn_energy_cutoff": self.intermediate_rxn_energy_cutoff,
            "use_basic_enumerator": self.use_basic_enumerator,
            "use_minimize_enumerator": self.use_minimize_enumerator,
            "filter_interdependent": self.filter_interdependent,
        }

        doc = PathwaySolverTaskDocument(**data)
        doc.task_label = self.name

        return doc



"""
Utility functions used by the enumerator classes.
"""
from typing import Dict, Iterable, List, Optional, Set, Tuple, Union

from pymatgen.analysis.phase_diagram import PhaseDiagram
from pymatgen.core.composition import Element
from pymatgen.entries.computed_entries import ComputedEntry, Entry

from rxn_network.core.reaction import Reaction
from rxn_network.entries.entry_set import GibbsEntrySet
from rxn_network.reactions.computed import ComputedReaction
from rxn_network.reactions.open import OpenComputedReaction


def get_elems_set(entries: Iterable[Entry]) -> Set[str]:
    """
    Returns chemical system as a set of element names, for set of entries.

    Args:
        entries: An iterable of entry-like objects

    Returns:
        Set of element names (strings).
    """
    return {str(elem) for e in entries for elem in e.composition.elements}


def get_total_chemsys_str(
    entries: Iterable[Entry], open_elems: Optional[Iterable[Union[Element]]] = None
) -> str:
    """
    Returns chemical system string for set of entries, with optional open element.

    Args:
        entries: An iterable of entry-like objects
        open_elem: optional open element to include in chemical system
    """
    elements = {elem for entry in entries for elem in entry.composition.elements}
    if open_elems:
        elements.update(list(open_elems))
    return "-".join(sorted([str(e) for e in elements]))


def group_by_chemsys(
    combos: Iterable[Tuple[Entry]], open_elems: Optional[Iterable[Element]] = None
) -> dict:
    """
    Groups entry combinations by chemical system, with optional open element.

    Args:
        combos: Iterable of entry combinations
        open_elem: optional open element to include in chemical system grouping

    Returns:
        Dictionary of entry combos grouped by chemical system
    """
    combo_dict: Dict[str, List[Tuple[Entry]]] = {}
    for combo in combos:
        key = get_total_chemsys_str(combo, open_elems)
        if key in combo_dict:
            combo_dict[key].append(combo)
        else:
            combo_dict[key] = [combo]

    return combo_dict


def stabilize_entries(
    pd: PhaseDiagram, entries_to_adjust: Iterable[Entry], tol: float = 1e-6
) -> List[Entry]:
    """
    Simple method for stabilizing entries by decreasing their energy to be on the hull.

    WARNING: This method is not guaranteed to work *simultaneously* for all entries due
    to the fact that stabilization of one entry may destabilize others. Use with
    caution.

    Args:
        pd: PhaseDiagram object
        entries_to_adjust: Iterable of entries requiring energies to be adjusted
        tol: Numerical tolerance to ensure that the energy of the entry is below the
            hull

    Returns:
        A list of new entries with energies adjusted to be on the hull
    """
    indices = [pd.all_entries.index(entry) for entry in entries_to_adjust]

    new_entries = []
    for _, entry in zip(indices, entries_to_adjust):
        e_above_hull = pd.get_e_above_hull(entry)
        entry_dict = entry.to_dict()
        entry_dict["energy"] = entry.uncorrected_energy + (
            e_above_hull * entry.composition.num_atoms - tol
        )
        new_entry = ComputedEntry.from_dict(entry_dict)
        new_entries.append(new_entry)

    return new_entries


def get_computed_rxn(
    rxn: Reaction, entries: GibbsEntrySet, chempots=None
) -> ComputedReaction:
    """
    Provided with a Reaction object and a list of possible entries, this function
    returns a new ComputedReaction object containing a selection of those entries.

    Args:
        rxn: Reaction object
        entries: Iterable of entries

    Returns:
        A ComputedReaction object transformed from a normal Reaction object
    """
    reactant_entries = [
        entries.get_min_entry_by_formula(r.reduced_formula) for r in rxn.reactants
    ]
    product_entries = [
        entries.get_min_entry_by_formula(p.reduced_formula) for p in rxn.products
    ]

    if chempots:
        rxn = OpenComputedReaction.balance(reactant_entries, product_entries, chempots)
    else:
        rxn = ComputedReaction.balance(reactant_entries, product_entries)

    return rxn


def get_rxn_info(rxn: ComputedReaction):
    """
    Utility function to get basic reaction information from a ComputedReaction object.
    Used in enumerators.

    Args:
        rxn: ComputedReaction object

    Returns:
        Tuple of reaction information (entry indices, coefficients, data)
    """
    return [e.data["idx"] for e in rxn.entries], list(rxn.coefficients), rxn.data



from dataclasses import dataclass, field
from typing import List, Optional

from jobflow import Flow, Maker
from pymatgen.core.composition import Element

from rxn_network.core.composition import Composition
from rxn_network.enumerators.basic import BasicEnumerator, BasicOpenEnumerator
from rxn_network.enumerators.minimize import (
    MinimizeGibbsEnumerator,
    MinimizeGrandPotentialEnumerator,
)
from rxn_network.jobs.core import (
    GetEntrySetMaker,
    NetworkMaker,
    PathwaySolverMaker,
    ReactionEnumerationMaker,
)
from rxn_network.utils.funcs import get_logger

logger = get_logger(__name__)


@dataclass
class NetworkFlowMaker(Maker):
    name: str = "find_reaction_pathways"
    get_entry_set_maker: GetEntrySetMaker = field(default_factory=GetEntrySetMaker)
    enumeration_maker: ReactionEnumerationMaker = field(
        default_factory=ReactionEnumerationMaker
    )
    network_maker: NetworkMaker = field(default_factory=NetworkMaker)
    solver_maker: Optional[PathwaySolverMaker] = None
    open_elem: Optional[Element] = None
    chempots: Optional[List[float]] = None
    use_basic_enumerators: bool = True
    use_minimize_enumerators: bool = True
    basic_enumerator_kwargs: dict = field(default_factory=dict)
    minimize_enumerator_kwargs: dict = field(default_factory=dict)

    def __post_init__(self):
        self.open_elem = Element(self.open_elem) if self.open_elem else None
        self.open_formula = (
            Composition(str(self.open_elem)).reduced_formula if self.open_elem else None
        )

    def make(self, precursors, targets, entries=None):
        precursor_formulas = [Composition(f).reduced_formula for f in precursors]
        target_formulas = [Composition(f).reduced_formula for f in targets]

        flow_name = (
            f"Reaction Network analysis: {'-'.join(sorted(precursor_formulas))} ->"
            f" {'-'.join(sorted(target_formulas))}"
        )
        chemsys = "-".join(
            {
                str(e)
                for formula in precursor_formulas + target_formulas
                for e in Composition(formula).elements
            }
        )

        jobs = []

        if entries is None:
            get_entry_set_maker = self.get_entry_set_maker.update_kwargs(
                {
                    "name": self.get_entry_set_maker.name
                    + f" ({chemsys}, T={self.get_entry_set_maker.temperature} K,"
                    f" +{round(self.get_entry_set_maker.e_above_hull, 3)} eV)",
                    "formulas_to_include": list(
                        set(
                            self.get_entry_set_maker.formulas_to_include
                            + precursor_formulas
                            + target_formulas
                        )
                    ),
                }
            )
            get_entry_set_job = get_entry_set_maker.make(chemsys)
            jobs.append(get_entry_set_job)
            entries = get_entry_set_job.output.entries

        self.basic_enumerator_kwargs.copy()
        self.minimize_enumerator_kwargs.copy()

        enumerators = []

        if self.use_basic_enumerators:
            enumerators.append(
                BasicEnumerator(
                    **self.basic_enumerator_kwargs,
                )
            )
            if self.open_formula:
                enumerators.append(
                    BasicOpenEnumerator(
                        open_phases=[self.open_formula],
                        **self.basic_enumerator_kwargs,
                    )
                )
        if self.use_minimize_enumerators:
            enumerators.append(
                MinimizeGibbsEnumerator(
                    **self.minimize_enumerator_kwargs,
                )
            )

        enumeration_job = self.enumeration_maker.make(
            enumerators=enumerators, entries=entries
        )
        jobs.append(enumeration_job)

        base_rxn_set = enumeration_job.output.rxns

        base_network_job = self.network_maker.make([base_rxn_set])
        jobs.append(base_network_job)

        if self.solver_maker:
            base_pathway_job = self.solver_maker.make(
                base_network_job.output.paths,
                entries=base_network_job.output.network.entries,
            )
            jobs.append(base_pathway_job)

        if self.use_minimize_enumerators and self.open_elem and self.chempots:
            for chempot in self.chempots:
                subname = f"(open {str(self.open_elem)}, mu={chempot})"
                enumeration_maker = self.enumeration_maker.update_kwargs(
                    {"name": self.enumeration_maker.name + subname},
                    nested=False,
                )
                network_maker = self.network_maker.update_kwargs(
                    {
                        "name": self.network_maker.name + subname,
                        "chempot": chempot,
                        "open_elem": self.open_elem,
                    },
                    nested=False,
                )
                if self.solver_maker:
                    solver_maker = self.solver_maker.update_kwargs(
                        {
                            "name": self.solver_maker.name + subname,
                            "chempot": chempot,
                            "open_elem": self.open_elem,
                        },
                        nested=False,
                    )

                enumerator = MinimizeGrandPotentialEnumerator(
                    open_elem=self.open_elem,
                    mu=chempot,
                    **self.minimize_enumerator_kwargs,
                )
                enumeration_job = enumeration_maker.make(
                    enumerators=[enumerator], entries=entries
                )
                network_job = network_maker.make(
                    [base_rxn_set, enumeration_job.output.rxns]
                )
                jobs.extend([enumeration_job, network_job])
                if self.solver_maker:
                    pathway_job = solver_maker.make(
                        network_job.output.paths,
                        entries=network_job.output.network.entries,
                    )
                    jobs.append(pathway_job)

        return Flow(jobs, name=flow_name)



"""
Implements an Entry that looks up pre-tabulated Gibbs free energies from the NIST-JANAF
tables.
"""
from typing import Dict, List, Optional

from pymatgen.entries.computed_entries import EnergyAdjustment

from rxn_network.core.composition import Composition
from rxn_network.data import PATH_TO_NIST, load_experimental_data
from rxn_network.entries.experimental import ExperimentalReferenceEntry

G_COMPOUNDS = load_experimental_data(PATH_TO_NIST / "compounds.json.gz")


class NISTReferenceEntry(ExperimentalReferenceEntry):
    """
    An Entry class for NIST-JANAF experimental reference data. Given a composition,
    automatically finds the Gibbs free energy of formation, dGf(T) from tabulated
    reference values.

    Reference:
        Malcolm W. Chase Jr. NIST-JANAF thermochemical tables. Fourth edition.
        Washington, DC : American Chemical Society;  New York : American Institute of
        Physics for the National Institute of Standards and Technology, 1998.
    """

    REFERENCES = G_COMPOUNDS

    def __init__(
        self,
        composition: Composition,
        temperature: float,
        energy_adjustments: Optional[List[EnergyAdjustment]] = None,
        data: Optional[Dict] = None,
    ):
        """
        Args:
            composition: Composition object (within pymatgen).
            temperature: Temperature in Kelvin. If temperature is not selected from
                one of [300, 400, 500, ... 2000 K], then free energies will be
                interpolated. Defaults to 300 K.
            data: Optional dictionary containing entry data
        """
        super().__init__(
            composition=composition,
            temperature=temperature,
            energy_adjustments=energy_adjustments,
            data=data,
        )



"""
Basic interface for a reaction cost Calculator
"""
from abc import ABCMeta, abstractmethod
from copy import deepcopy
from typing import List

from monty.json import MSONable

from rxn_network.core.reaction import Reaction


class Calculator(MSONable, metaclass=ABCMeta):
    """
    Base definition for a property calculator
    """

    @abstractmethod
    def calculate(self, rxn: Reaction) -> float:
        """
        Evaluates the specified property of a reaction
        """

    def decorate(self, rxn: Reaction) -> Reaction:
        """
        Returns a copy of the reaction with the calculated property by
        storing the value within the reaction's data dictionary.

        Args:
            rxn: The reaction object.

        Returns:
            The reaction object, modified in place
        """
        new_rxn = deepcopy(rxn)

        if not getattr(new_rxn, "data"):
            new_rxn.data = {}

        new_rxn.data[self.name] = self.calculate(new_rxn)
        return new_rxn

    def calculate_many(self, rxns: List[Reaction]) -> List[float]:
        """
        Calculates the competitiveness score for a list of reactions by enumerating
        competing reactions, evaluating their cost with the supplied cost function, and
        then using the c-score formula, i.e. the _get_c_score() method, to determine the
        competitiveness score. Parallelized with ray.

        Args:
            rxns: the list of ComputedReaction objects to be evaluated

        Returns:
            The list of competitiveness scores
        """
        results = []
        for rxn in rxns:
            results.append(self.calculate(rxn))

        return results

    def decorate_many(self, rxns: List[Reaction]) -> List[Reaction]:
        """
        Decorates a list of reactions with the calculated properties.

        Args:
            rxns: the list of ComputedReaction objects to be decorated

        Returns:
            The list of decorated ComputedReaction objects
        """
        new_rxns = []
        for rxn in rxns:
            new_rxns.append(self.decorate(rxn))
        return new_rxns

    @property
    @abstractmethod
    def name(self):
        """
        The name of the calculator; used to store the value within the reaction's data
        dictionary
        """



"""
This module implements two types of reaction enumerators using a free energy
minimization technique, with or without the option of an open entry.
"""

from itertools import product
from typing import List, Optional

from pymatgen.analysis.interface_reactions import (
    GrandPotentialInterfacialReactivity,
    InterfacialReactivity,
)
from pymatgen.core.composition import Element

from rxn_network.core.composition import Composition
from rxn_network.enumerators.basic import BasicEnumerator
from rxn_network.enumerators.utils import get_computed_rxn


class MinimizeGibbsEnumerator(BasicEnumerator):
    """
    Enumerator for finding all reactions between two reactants that are predicted by
    thermodynamics; i.e., they appear when taking the convex hull along a straight line
    connecting any two phases in G-x phase space. Identity reactions are automatically
    excluded.
    """

    CHUNK_SIZE = 10000

    def __init__(
        self,
        precursors: Optional[List[str]] = None,
        targets: Optional[List[str]] = None,
        exclusive_precursors: bool = True,
        exclusive_targets: bool = False,
        filter_by_chemsys: Optional[str] = None,
        max_num_constraints: int = 1,
        calculate_e_above_hulls: bool = False,
        quiet: bool = False,
    ):
        """
        Args:
            precursors: Optional formulas of precursors.
            targets: Optional formulas of targets; only reactions which make
                these targets will be enumerated.
            calculators: Optional list of Calculator object names; see calculators
                module for options (e.g., ["ChempotDistanceCalculator"])
            exclusive_precursors: Whether to consider only reactions that have
                reactants which are a subset of the provided list of precursors.
                Defaults to True.
            exclusive_targets: Whether to consider only reactions that make the
                provided target directly (i.e. with no byproducts). Defualts to False.
            quiet: Whether to run in quiet mode (no progress bar). Defaults to False.
        """
        super().__init__(
            precursors=precursors,
            targets=targets,
            exclusive_precursors=exclusive_precursors,
            exclusive_targets=exclusive_targets,
            filter_by_chemsys=filter_by_chemsys,
            quiet=quiet,
        )
        self._build_pd = True

    @staticmethod
    def _react_function(
        reactants, products, filtered_entries=None, pd=None, grand_pd=None, **kwargs
    ):
        """React method for MinimizeGibbsEnumerator, which uses the interfacial reaction
        approach (see _react_interface())"""

        r = list(reactants)
        r0 = r[0]

        if len(r) == 1:
            r1 = r[0]
        else:
            r1 = r[1]

        return react_interface(
            r0.composition,
            r1.composition,
            filtered_entries,
            pd,
            grand_pd,
        )

    @staticmethod
    def _get_rxn_iterable(combos, open_combos):
        """Gets the iterable used to generate reactions"""
        _ = open_combos  # unused argument

        return product(combos, [None])

    @staticmethod
    def _rxn_iter_length(combos, open_combos):
        _ = open_combos
        return len(combos)


class MinimizeGrandPotentialEnumerator(MinimizeGibbsEnumerator):
    """
    Enumerator for finding all reactions between two reactants and an open element
    that are predicted by thermo; i.e., they appear when taking the
    convex hull along a straight line connecting any two phases in Phi-x
    phase space. Identity reactions are excluded.
    """

    CHUNK_SIZE = 10000

    def __init__(
        self,
        open_elem: Element,
        mu: float,
        precursors: Optional[List[str]] = None,
        targets: Optional[List[str]] = None,
        exclusive_precursors: bool = True,
        exclusive_targets: bool = False,
        filter_by_chemsys: Optional[str] = None,
        max_num_constraints=1,
        quiet: bool = False,
    ):
        """
        Args:
            open_elem: The element to be considered as open
            mu: The chemical potential of the open element
            precursors: Optional formulas of precursors.
            targets: Optional formulas of targets; only reactions which make
                these targets will be enumerated.
            calculators: Optional list of Calculator object names; see calculators
                module for options (e.g., ["ChempotDistanceCalculator])
            exclusive_precursors: Whether to consider only reactions that have
                reactants which are a subset of the provided list of precursors.
                Defaults to True.
            exclusive_targets: Whether to consider only reactions that make the
                provided target directly (i.e. with no byproducts). Defualts to False.
            quiet: Whether to run in quiet mode (no progress bar). Defaults to False.
        """

        super().__init__(
            precursors=precursors,
            targets=targets,
            exclusive_precursors=exclusive_precursors,
            exclusive_targets=exclusive_targets,
            filter_by_chemsys=filter_by_chemsys,
            quiet=quiet,
        )
        self.open_elem = Element(open_elem)
        self.open_phases = [Composition(str(self.open_elem)).reduced_formula]
        self.mu = mu
        self.chempots = {self.open_elem: self.mu}
        self._build_grand_pd = True

    @staticmethod
    def _react_function(
        reactants, products, filtered_entries=None, pd=None, grand_pd=None, **kwargs
    ):
        """Same as the MinimizeGibbsEnumerator react function, but with ability to
        specify open element and grand potential phase diagram"""
        r = list(reactants)
        r0 = r[0]

        if len(r) == 1:
            r1 = r[0]
        else:
            r1 = r[1]

        open_elem = list(grand_pd.chempots.keys())[0]

        for reactant in r:
            elems = reactant.composition.elements
            if len(elems) == 1 and elems[0] == open_elem:  # skip if reactant = open_e
                return []

        return react_interface(
            r0.composition,
            r1.composition,
            filtered_entries,
            pd,
            grand_pd=grand_pd,
        )


def react_interface(r1, r2, filtered_entries, pd, grand_pd=None):
    """Simple API for InterfacialReactivity module from pymatgen."""
    chempots = None

    if grand_pd:
        interface = GrandPotentialInterfacialReactivity(
            r1,
            r2,
            grand_pd,
            pd_non_grand=pd,
            norm=True,
            include_no_mixing_energy=True,
            use_hull_energy=True,
        )
        chempots = grand_pd.chempots

    else:
        interface = InterfacialReactivity(
            r1,
            r2,
            pd,
            use_hull_energy=True,
        )

    rxns = []
    for _, _, _, rxn, _ in interface.get_kinks():
        rxn = get_computed_rxn(rxn, filtered_entries, chempots)
        rxns.append(rxn)

    return rxns



"""Code for analyzing sets of reactions between two phases."""

from functools import cached_property
from typing import List

import numpy as np
import plotly.express as px
from monty.json import MSONable
from plotly.graph_objs import Figure
from scipy.spatial import ConvexHull

from rxn_network.core.composition import Composition
from rxn_network.reactions.computed import ComputedReaction


class InterfaceReactionHull(MSONable):
    """
    A class for storing and analyzing a set of reactions at an interface between two
    reactants. This class is more generalized than the InterfacialReactivity class and
    can encompass any set of reactions between two reactants, regardless of whether
    the reaction products are "stable" (i.e. together on the convex hull)
    """

    def __init__(
        self,
        c1: Composition,
        c2: Composition,
        reactions: List[ComputedReaction],
        max_num_constraints=1,
    ):
        """
        Args:
            c1: Composition of reactant 1
            c2: Composition of reactant 2
            reactions: List of reactions containing all enumerated reactions between the
                two reactants. Note that this list should
        """
        self.c1 = Composition(c1).reduced_composition
        self.c2 = Composition(c2).reduced_composition
        self.e1 = None
        self.e2 = None

        for rxn in reactions:
            for e in rxn.reactant_entries:
                if e.composition.reduced_composition == self.c1:
                    self.e1 = e
                elif e.composition.reduced_composition == self.c2:
                    self.e2 = e
            if self.e1 is not None and self.e2 is not None:
                break
        else:
            print(c1, c2)
            print(reactions)
            raise ValueError(
                "Provided reactions do not correspond to reactant compositons!"
            )

        reactions = [
            r
            for r in reactions
            if r.data.get("num_constraints", 1) <= max_num_constraints
        ]

        endpoint_reactions = [
            ComputedReaction.balance([self.e1], [self.e1]),
            ComputedReaction.balance([self.e2], [self.e2]),
        ]

        reactions_with_endpoints = reactions + endpoint_reactions

        coords = np.array(
            [(self.get_coordinate(r), r.energy_per_atom) for r in reactions]
        )
        coords = np.append(coords, [[0, 0], [1, 0]], axis=0)

        idx_sort = coords[:, 0].argsort()

        self.coords = coords[idx_sort]
        self.reactions = [reactions_with_endpoints[i] for i in idx_sort]
        self.hull = ConvexHull(self.coords)
        self.endpoint_reactions = endpoint_reactions

    def plot(self, y_max=0.2):
        """
        Plot the reaction hull.
        """
        pts = self._get_scatter()
        lines = self._get_lines()

        fig = Figure(data=lines + [pts])

        fig.update_traces(
            hovertemplate=(
                "<b>%{hovertext}</b><br> <br><b>Atomic fraction</b>:"
                " %{x:.3f}<br><b>Energy</b>: %{y:.3f} (eV/atom)"
            )
        )
        fig.update_layout(yaxis_range=[min(self.coords[:, 1]) - 0.01, y_max])
        fig.update_layout(xaxis_title="Mixing ratio")
        fig.update_layout(yaxis_title="Energy (eV/atom)")

        return fig

    def get_energy_above_hull(self, reaction):
        """
        Get the energy of a reaction above the reaction hull.
        """
        idx = self.reactions.index(reaction)
        x, y = self.coords[idx]
        e_above_hull = y - self.get_hull_energy(x)

        return e_above_hull

    def get_coordinate(self, reaction):
        """Get coordinate of reaction in reaction hull. This is expressed as the atomic
        mixing ratio of component 2 in the reaction."""
        amt_c1 = reaction.reactant_atomic_fractions.get(self.c1, 0)
        amt_c2 = reaction.reactant_atomic_fractions.get(self.c2, 0)
        total = amt_c1 + amt_c2  # will add to 1.0 with two-component reactions
        try:
            coordinate = amt_c2 / total
        except ZeroDivisionError as e:
            raise ValueError(
                f"Can't compute coordinate for {reaction} with {self.c1}, {self.c2}"
            ) from e

        return round(coordinate, 12)  # avoids numerical issues

    def get_hull_energy(self, coordinate):
        """
        Get the energy of the reaction at a given coordinate.

        Args:
            coordinate: Coordinate of reaction in reaction hull.

        Returns:
            Energy of reaction at given coordinate.
        """
        reactions = self.get_reactions_by_coordinate(coordinate)
        return sum(weight * r.energy_per_atom for r, weight in reactions.items())

    def get_reactions_by_coordinate(self, coordinate):
        """Get the reaction(s) at a given coordinate."""
        sorted_vertices = np.sort(self.hull_vertices)
        for i in range(len(sorted_vertices) - 1):
            v1 = sorted_vertices[i]
            v2 = sorted_vertices[i + 1]

            x1 = self.coords[v1, 0]
            x2 = self.coords[v2, 0]

            if np.isclose(coordinate, x1):
                return {self.reactions[v1]: 1.0}
            if np.isclose(coordinate, x2):
                return {self.reactions[v2]: 1.0}
            if x1 < coordinate < x2:
                return {
                    self.reactions[v1]: (x2 - coordinate) / (x2 - x1),
                    self.reactions[v2]: (coordinate - x1) / (x2 - x1),
                }

            continue

        raise ValueError("No reactions found!")

    def get_coords_in_range(self, x1, x2):
        """
        Get the coordinates in the range [x1, x2].

        Args:
            x1: Start of range.
            x2: End of range.

        Returns:
            Array of coordinates in the range.

        """
        x_min, x_max = sorted([x1, x2])
        y_min, y_max = self.get_hull_energy(x_min), self.get_hull_energy(x_max)

        coords = []

        if np.isclose(x_min, 0.0) and not np.isclose(y_min, 0.0):
            coords.append([0.0, 0.0])

        coords.append([x_min, y_min])

        coords.extend(
            [
                self.coords[i]
                for i in self.hull_vertices
                if self.coords[i, 0] < x_max
                and self.coords[i, 0] > x_min
                and self.coords[i, 1] <= 0
            ]
        )

        if x_max != x_min:
            coords.append([x_max, y_max])
        if np.isclose(x_max, 1.0) and not np.isclose(y_max, 0.0):
            coords.append([1.0, 0.0])

        coords = np.array(coords)

        return coords[coords[:, 0].argsort()]

    def _get_scatter(self):
        marker_size = 10

        pts = px.scatter(
            x=self.coords[:, 0],
            y=self.coords[:, 1],
            hover_name=[str(r) for r in self.reactions],
            labels={
                "x": "Mixing Ratio",
                "y": (
                    r"$\Delta G_{\mathrm{rxn}} ~"
                    r" \mathrm{\left(\dfrac{\mathsf{eV}}{\mathsf{atom}}\right)}$"
                ),
            },
        )
        pts.update_traces(marker={"size": marker_size})

        return pts.data[0]

    def _get_lines(self):
        coords = self.coords[self.hull.simplices]

        coords = coords[(coords[:, :, 1] <= 0).all(axis=1)]
        coords = coords[~(coords[:, :, 1] == 0).all(axis=1)]

        lines = [
            px.line(x=c[:, 0], y=c[:, 1]) for c in coords if not (c[:, 1] == 0).all()
        ]

        line_data = []
        for line in lines:
            line.update_traces(line={"color": "black"})
            line_data.append(line.data[0])

        return line_data

    @cached_property
    def hull_vertices(self):
        hull_vertices = [
            i
            for i in self.hull.vertices
            if self.coords[i, 1] <= 0
            and np.isclose(
                self.coords[self.coords[:, 0] == self.coords[i, 0]][:, 1].min(),
                self.coords[i, 1],  # make sure point is lower than others on same x
            )
        ]

        return np.array(hull_vertices)

    @cached_property
    def stable_reactions(self):
        """
        Returns the reactions that are stable (on the convex hull) of the interface
        reaction hull.
        """
        return [r for i, r in enumerate(self.reactions) if i in self.hull_vertices]

    @cached_property
    def unstable_reactions(self):
        """
        Returns the reactions that are unstable (NOT on the convex hull) of the
        interface reaction hull.
        """
        return [r for i, r in enumerate(self.reactions) if i not in self.hull_vertices]



"""
Basic interface for a (chemical) Reaction
"""
from abc import ABCMeta, abstractmethod
from typing import List

import numpy as np
from monty.json import MSONable
from pymatgen.core.composition import Element

from rxn_network.core.composition import Composition


class Reaction(MSONable, metaclass=ABCMeta):
    """
    Base definition for a reaction class.
    """

    @property
    @abstractmethod
    def reactants(self) -> List[Composition]:
        """List of reactants for this reaction"""

    @property
    @abstractmethod
    def products(self) -> List[Composition]:
        """List of products for this reaction"""

    @property
    @abstractmethod
    def coefficients(self) -> np.ndarray:
        """Coefficients of the reaction"""

    @property
    @abstractmethod
    def energy(self) -> float:
        """The energy of this reaction in total eV"""

    @property
    @abstractmethod
    def compositions(self) -> List[Composition]:
        """List of all compositions in the reaction"""

    @property
    def elements(self) -> List[Element]:
        """List of elements in the reaction"""
        return list(set(el for comp in self.compositions for el in comp.elements))



"""
A calculator class for determining chemical potential distance of reactions
"""
from itertools import chain, combinations, product
from typing import Callable, List

import numpy as np
from pymatgen.analysis.phase_diagram import PDEntry

from rxn_network.core.calculator import Calculator
from rxn_network.reactions.computed import ComputedReaction
from rxn_network.thermo.chempot_diagram import ChemicalPotentialDiagram


class ChempotDistanceCalculator(Calculator):
    """
    Calculator for determining the "chemical potential distance" for a reaction
    (in eV/atom).

    For more information on this specific implementation of the algorithm,
    please reference the paper below:

    Todd, P. K.; McDermott, M. J.; Rom, C. L.; Corrao, A. A.; Denney, J. J.; Dwaraknath,
    S. S.; Khalifah, P. G.; Persson, K. A.;  Neilson, J. R. Selectivity in Yttrium
    Manganese Oxide Synthesis via Local Chemical Potentials in Hyperdimensional Phase
    Space. J. Am. Chem. Soc. 2021, 143 (37), 15185–15194.
    https://doi.org/10.1021/jacs.1c06229.

    """

    def __init__(
        self,
        cpd: ChemicalPotentialDiagram,
        mu_func: str = "sum",
        name: str = "chempot_distance",
    ):
        """
        Args:
            cpd: the chemical potential diagram object
            mu_func: the name of the function used to process the interfacial
                chemical potential distances into a single value describing the whole
                reaction. Current options are 1) max, 2) mean, and 3) sum (default).
            name: the data dictionary key with which to store the calculated value.
                Defaults to "chempot_distance".
        """
        self.cpd = cpd
        self._name = name

        if mu_func == "max":
            self._mu_func = max  # type: ignore
        elif mu_func == "mean":
            self._mu_func = np.mean  # type: ignore
        elif mu_func == "sum":
            self._mu_func = sum  # type: ignore

        self._open_elems = set()
        if cpd.entries[0].__class__.__name__ == "GrandPotPDEntry":
            self._open_elems = set(cpd.entries[0].chempots.keys())

    def calculate(self, rxn: ComputedReaction) -> float:
        """
        Calculates the (aggregate) chemical potential distance in eV/atom. The mu_func
        parameter determines how the individual pairwise interface distances are
        aggregated into a single value describing the overall reaction.

        Args:
            rxn: the ComputedReaction object

        Returns:
            The chemical potential distance of the reaction.
        """
        reactant_entries = rxn.reactant_entries
        product_entries = rxn.product_entries

        if hasattr(rxn, "grand_entries"):
            reactant_entries = [
                e
                for e, c in zip(rxn.grand_entries, rxn.coefficients)
                if c < 0 and e.__class__.__name__ == "GrandPotPDEntry"
            ]
            product_entries = [
                e
                for e, c in zip(rxn.grand_entries, rxn.coefficients)
                if c > 0 and e.__class__.__name__ == "GrandPotPDEntry"
            ]
        combos = chain(
            product(reactant_entries, product_entries),
            combinations(product_entries, 2),
        )
        distances = [
            self.cpd.shortest_domain_distance(
                combo[0].composition.reduced_formula,
                combo[1].composition.reduced_formula,
                offset=self.cpd.get_offset(combo[0]) + self.cpd.get_offset(combo[1]),
            )
            for combo in combos
        ]

        distance = round(float(self._mu_func(distances)), 5)
        return distance

    @classmethod
    def from_entries(
        cls,
        entries: List[PDEntry],
        mu_func: str = "sum",
        name: str = "chempot_distance",
        **kwargs,
    ) -> "ChempotDistanceCalculator":
        """
        Convenience constructor which first builds the ChemicalPotentialDiagram
        object from a list of entry objects.

        Args:
            entries: entry objects used to build the ChemicalPotentialDiagram
            mu_func: the name of the function used to process the interfacial
                chemical potential distances into a single value describing the whole
                reaction.
            name: the data dictionary key by which to store the calculated value,
                defaults to "chempot_distance"
            **kwargs: optional kwargs passed to ChemicalPotentialDiagram.

        Returns:
            A ChempotDistanceCalculator object
        """

        cpd = ChemicalPotentialDiagram(entries=entries, **kwargs)
        return cls(cpd, mu_func, name)

    @property
    def mu_func(self) -> Callable:
        """Returns the function used to process the chemical potential distances into a
        single metric."""
        return self._mu_func

    @property
    def name(self) -> str:
        """Returns the name of the data dictionary key where the value is stored"""
        return self._name



""" 
Script for generating file: data/nist/compounds.json.gz
"""

import gzip
import json
from collections import OrderedDict
from typing import Dict

import numpy as np
import pandas
from thermochem.janaf import Janafdb
from tqdm import tqdm

from rxn_network.core.composition import Composition

JMOL_PER_EV = 96485.307499258

if __name__ == "__main__":
    jdb = Janafdb()

    db = pandas.read_csv(
        "JANAF_index.txt",
        delimiter="|",
    )
    db.columns = ["formula", "name", "phase", "filename"]
    db["formula"] = db["formula"].map(str.strip)
    db["name"] = db["name"].map(str.strip)
    db["phase"] = db["phase"].map(str.strip)
    db["filename"] = db["filename"].map(str.strip)

    # Clean formulas

    cleaned_formulas = []
    for f in db["formula"].unique():
        if "-" in f or "+" in f or "D" in f:  # remove ions
            continue

        comp = Composition(f)
        if comp.is_element:
            continue

        cleaned_formulas.append(f)

    # Acquire data
    all_data: Dict[str, Dict[float, float]] = {}

    for f in tqdm(cleaned_formulas):
        phase_df = db[db["formula"] == f]
        phases = phase_df["phase"]

        # Priortize mixed phases and then single phases
        if "cr,l" in phases.unique():
            phase = phase_df[phases == "cr,l"]
        elif "l,g" in phases.unique():
            phase = phase_df[phases == "l,g"]
        elif "cr" in phases.unique():
            phase = phase_df[phases == "cr"]
        elif "l" in phases.unique():
            phase = phase_df[phases == "l"]
        elif "g" in phases.unique():
            phase = phase_df[phases == "g"]

        fn = phase["filename"].values[0]

        try:
            data = jdb.getphasedata(filename=fn)
        except pandas.errors.ParserError as e:
            print(e)
            continue

        max_temp = int((data.rawdata["T"].max() // 100) * 100)
        max_temp = min(max_temp, 2000)
        temps = [i for i in range(300, max_temp + 100, 100)]

        dg_f = [data.DeltaG(t) / JMOL_PER_EV for t in temps]

        formula, factor = Composition(f).get_integer_formula_and_factor()
        energies = {temp: g / factor for temp, g in zip(temps, dg_f)}

        existing_data = all_data.get(formula)
        if not existing_data:
            new_energies = energies
        else:
            new_energies = dict()
            for t in sorted(set(temps + list(existing_data.keys()))):
                new_e = energies.get(t, np.inf)
                old_e = existing_data.get(t, np.inf)

                new_energies[t] = min(new_e, old_e)

        all_data[formula] = new_energies

    all_data = OrderedDict(sorted(all_data.items()))

    with gzip.open("compounds.json.gz", "wt", encoding="ascii") as zipfile:
        json.dump(all_data, zipfile)



"""
Basic interface for a reaction pathway.
"""
from abc import ABCMeta
from typing import List

from monty.json import MSONable

from rxn_network.core.reaction import Reaction


class Pathway(MSONable, metaclass=ABCMeta):
    """
    Base definition for a reaction pathway.
    """

    _reactions: List[Reaction]

    @property
    def entries(self):
        """Entry objects in this Pathway"""
        return {entry for rxn in self._reactions for entry in rxn.entries}

    @property
    def all_reactants(self):
        """Entries serving as a reactant in any sub reaction"""
        return {entry for rxn in self._reactions for entry in rxn.reactants}

    @property
    def all_products(self):
        """Entries serving as a product in any sub reaction"""
        return {entry for rxn in self._reactions for entry in rxn.products}

    @property
    def compositions(self):
        """Compositions in the reaction"""
        return list(self.all_reactants | self.all_products)

    @property
    def reactants(self):
        """The reactants of this whole reaction pathway"""
        return self.all_reactants - self.all_products

    @property
    def products(self):
        """The products of this whole reaction pathway"""
        return self.all_products - self.all_reactants

    @property
    def intermediates(self):
        """Intermediates as entries in this reaction pathway"""
        return self.all_products & self.all_reactants

    @property
    def energy(self):
        """Total energy of this reaction pathway"""
        return sum(rxn.energy for rxn in self._reactions)

    @property
    def energy_per_atom(self):
        """Total energy per atom of this reaction pathway"""
        return sum(rxn.energy_per_atom for rxn in self._reactions)



"""Definitions of common job functions"""

from pymatgen.core.composition import Element

from rxn_network.core.composition import Composition
from rxn_network.utils.funcs import get_logger

logger = get_logger(__name__)


def run_enumerators(enumerators, entries):
    rxn_set = None
    for enumerator in enumerators:
        logger.info(f"Running {enumerator.__class__.__name__}")
        rxns = enumerator.enumerate(entries)

        logger.info(f"Adding {len(rxns)} reactions to reaction set")

        if rxn_set is None:
            rxn_set = rxns
        else:
            rxn_set = rxn_set.add_rxn_set(rxns)

    logger.info("Completed reaction enumeration. Filtering duplicates...")
    rxn_set = rxn_set.filter_duplicates()
    return rxn_set


def get_added_elem_data(entries, targets):
    added_elems = entries.chemsys - {
        str(e) for target in targets for e in Composition(target).elements
    }
    added_chemsys = "-".join(sorted(list(added_elems)))
    added_elements = [Element(e) for e in added_elems]

    return added_elements, added_chemsys



"""
Basic interface for a reaction Enumerator
"""
import logging
from abc import ABCMeta, abstractmethod

from monty.json import MSONable


class Enumerator(MSONable, metaclass=ABCMeta):
    """
    Base definition for a class that enumerates reactions.
    """

    def __init__(self, precursors, targets):
        self.logger = logging.getLogger(str(self.__class__.__name__))
        self.logger.setLevel("INFO")
        self.precursors = precursors if precursors else []
        self.targets = targets if targets else []

    @abstractmethod
    def enumerate(self, entries):
        """
        Enumerates the potential reactions from the list of entries
        """
